/* support code for expat */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <errno.h>
#include "xml2he5.h"
#include "globals.h"

handlerstack *hstop = NULL;
handlerstack *hsold = NULL;

handlerstack *new_hs(void)
{
  handlerstack *result;
  if (hsold == NULL)
    result = (handlerstack *) malloc(sizeof(handlerstack));
  else {
    result = hsold;
    hsold = hsold->next;
  }
  result->eh = NULL;
  return result;
}

void recycle_hs(handlerstack *foo)
{
  foo->next = hsold;
  hsold = foo;
}

int hs_length(void)
{
  int len = 0;
  handlerstack *htmp = hstop;
  while(htmp != NULL) {
    len++;
    htmp = htmp->next;
  }
  return len;
}

/* because there is no published way to GetElementHandler, the convention
   is that the current handler values are in the node at the top of
   the stack.
*/
void push_end_handler(XML_Parser p, XML_EndElementHandler eh, const char *ehname)
{
  handlerstack *new = new_hs();
  
  XML_SetEndElementHandler(p, eh);
  new->eh = eh;
  new->name = ehname;
  new->next = hstop;
  new->elementdata = NULL;
  hstop = new;
  fprintf(stderr, "Stack depth: %d\n", hs_length());
}

void *pop_end_handler(XML_Parser p)
{
  void *result = NULL;

  if(hstop != NULL) {
    handlerstack *now = hstop;
    result = now->elementdata;
    hstop = hstop->next;
    recycle_hs(now);
 
    if(hstop != NULL)
      XML_SetEndElementHandler(p, hstop->eh);
  }
  fprintf(stderr, "Stack depth: %d\n", hs_length());
  return result;
}

void show_end_handler_stack(void)
{
  handlerstack *tmp = hstop;
  int i = 0;
  fprintf(stderr, "END HANDLER STACK:\n");
  while (tmp != NULL) {
    fprintf(stderr, "%2d: %s\n", i++, tmp->name);
    tmp = tmp->next;
  }
}

void set_element_data(void *data)
{
  if(hstop != NULL) {
    hstop->elementdata = data;
  }
  else {
    fprintf(stderr, "set_element_data on an empty stack!!\n");
    exit(-1);
  }
}

void *get_element_data(void)
{
  void *result = NULL;
  if(hstop != NULL) {
    result = hstop->elementdata;
  }
  else {
    fprintf(stderr, "set_element_data on an empty stack!!\n");
    exit(-1);
  }
  return result;
}

void set_end_handler(XML_Parser p)
{
  XML_SetEndElementHandler(p, hstop->eh);
}

/* This used to use userData to pass the address of the CDATA buffer.
   I took it out because something was randomly clobbering the global
   userData register.
*/


/* a call back function to collect CDATA fields */
/* the userData is a cdatabuf * */
void collect_cdata(void *userData, const XML_Char *s, int len)
{
  cdatabuf *cdb = &GlobalCDATA;

  if(cdb->current + len >= cdb->size) {
    unsigned long newsize = cdb->current + len + 16;
    char *newbuf = realloc(cdb->buffer, newsize);
    if(newbuf == NULL) {
      fprintf(stderr, "Ran out of memory resizing buffer to %ld ?!\n", newsize);
      exit(-1);
    }
    cdb->buffer = newbuf;
    cdb->size = newsize;
  }
  strncat(cdb->buffer, s, (unsigned)len);
  cdb->current += len;
  fprintf(stderr, "String value[%ld]: |%s|\n", cdb->current, cdb->buffer);
}

void start_collecting_cdata(XML_Parser p, cdatabuf *cdb)
{
  XML_SetCharacterDataHandler(p, collect_cdata);
  cdb->current = 0;
  cdb->buffer[0] = 0;
  /* XML_SetUserData(p, cdb); */
}

void stop_collecting_cdata(XML_Parser p)
{
  XML_SetCharacterDataHandler(p, NULL);
}

/* Simple attribute lookup */

XML_Char const *find_attribute(const XML_Char **atts, char const *aname)
{
  for( ; *atts != NULL; atts += 2)
    if(!strcmp(aname, *atts)) return *(atts + 1);

  return NULL;
}

/* This mess supports arbitrarily long CDATA containing white-space
   delimited numbers (signed/unsigned integers or float/doubles)
*/

/* The buffer into which the numbers get stored */

void *inline_databuf;
long inline_datacount;
long inline_dataindex;

/* pointers to stub of previous buffer (necessitated by the chunk-at-a-time
   semantics of expat CDATA callbacks
*/
char* previous_buffer;
char* previous_tail;

/* static locals to hold found values */
long inline_long;
unsigned long inline_ulong;
double inline_double;

/* processing stub functions to be shoved into pointers above */

/* predeclarations to shut compiler up */
const char *process_double(const char *ctail);
void finish_double(void);
void finish_float(void);
const char *process_long(const char *ctail);
void finish_long(void);
void finish_int(void);
void finish_short(void);
void finish_char(void);
const char *process_ulong(const char *ctail);
void finish_ulong(void);
void finish_uint(void);
void finish_ushort(void);
void finish_uchar(void);

/* processing for double and float */

const char *process_double(const char *ctail)
{
  char *newtail;

  inline_double = strtod(ctail, &newtail);

  /* Check for errors, and spit appropriately */

  if(newtail == ctail
     || ((inline_double == HUGE_VAL
	  || inline_double == -HUGE_VAL
	  || inline_double == 0.0)
	 && errno == ERANGE))
    return ctail;
  else
    return newtail;
}

void finish_double(void)
{
  ((double *)inline_databuf)[inline_dataindex++] = inline_double;
  inline_datacount--;
}

void finish_float(void)
{
  ((float *)inline_databuf)[inline_dataindex++] = inline_double;
  inline_datacount--;
}

/* processing for signed long/int/char */

const char *process_long(const char *ctail)
{
  char *newtail;

  inline_long = strtol(ctail, &newtail, 10);

  /* Check for errors, and spit appropriately */

  if(newtail == ctail
     || ((inline_double == LONG_MIN
	  || inline_double == LONG_MAX)
	 && errno == ERANGE))
    return ctail;
  else
    return newtail;
}

void finish_long(void)
{
  ((long *)inline_databuf)[inline_dataindex++] = inline_long;
  inline_datacount--;
}

void finish_int(void)
{
  ((int *)inline_databuf)[inline_dataindex++] = inline_long;
  inline_datacount--;
}

void finish_short(void)
{
  ((short *)inline_databuf)[inline_dataindex++] = inline_long;
  inline_datacount--;
}

void finish_char(void)
{
  ((char *)inline_databuf)[inline_dataindex++] = inline_long;
  inline_datacount--;
}

/* processing for unsigned long/int/char */

const char *process_ulong(const char *ctail)
{
  char *newtail;

  inline_ulong = strtoul(ctail, &newtail, 10);

  /* Check for errors, and spit appropriately */

  if(newtail == ctail
     || (inline_double == LONG_MIN && errno == ERANGE))
    return ctail;
  else
    return newtail;
}

void finish_ulong(void)
{
  ((unsigned long *)inline_databuf)[inline_dataindex++] = inline_ulong;
  inline_datacount--;
}

void finish_uint(void)
{
  ((unsigned int *)inline_databuf)[inline_dataindex++] = inline_ulong;
  inline_datacount--;
}

void finish_ushort(void)
{
  ((unsigned short *)inline_databuf)[inline_dataindex++] = inline_ulong;
  inline_datacount--;
}

void finish_uchar(void)
{
  ((unsigned char *)inline_databuf)[inline_dataindex++] = inline_ulong;
  inline_datacount--;
}
/* pointers to functions used by main buffer_munching loop */

const char *(*process_next)(const char *ctail);
void (*finish_processing)(void);

void inline_data_init(void *buf, long count, hid_t native_type )
{
  TRACE((stderr, "inline_data_init, %ld values\n", count));
  inline_databuf = buf;
  inline_datacount = count;
  inline_dataindex = 0;
  previous_buffer = NULL;
  previous_tail = NULL;
  /* Set up the processing functions according to the current
     datatype. */
  if (native_type ==  H5T_NATIVE_SCHAR) {
    process_next = process_long; finish_processing = finish_char;}
  else if (native_type ==  H5T_NATIVE_UCHAR) {
    process_next = process_ulong; finish_processing = finish_uchar;}
  else if (native_type ==  H5T_NATIVE_SHORT) {
    process_next = process_long; finish_processing = finish_short;}
  else if (native_type ==  H5T_NATIVE_USHORT) {
    process_next = process_ulong; finish_processing = finish_ushort;}
  else if (native_type ==  H5T_NATIVE_INT) {
    process_next = process_long; finish_processing = finish_int;}
  else if (native_type ==  H5T_NATIVE_UINT) {
    process_next = process_ulong; finish_processing = finish_uint;}
  else if (native_type ==  H5T_NATIVE_LLONG) {
    process_next = process_long; finish_processing = finish_long;}
  else if (native_type ==  H5T_NATIVE_ULLONG) {
    process_next = process_ulong; finish_processing = finish_ulong;}
  else if (native_type ==  H5T_NATIVE_FLOAT) {
    process_next = process_double; finish_processing = finish_float;}
  else if (native_type ==  H5T_NATIVE_DOUBLE) {
    process_next = process_double; finish_processing = finish_double;}
  else {
    fprintf(stderr, "Non-numeric type in in_line_data: %d\n",
	    native_type);
    PrintXMLLocation();
    exit(-1);
  }
  XML_SetCharacterDataHandler(TheParser, process_in_line_buffer);
}

/* and, finally, the actual buffer processor CDATA callback function */

void process_in_line_buffer(void *userData, const XML_Char *s, int len)
{
  char *current;
  char *ctail;
  char *cend;

  TRACE((stderr, "inline_buffer\n"));

  /* merge in previous data with current data */
  if(previous_buffer != NULL) {
    int newlength = len + strlen(previous_tail) + 1;

    current = newstring(newlength);
    strcat(current, previous_tail);
    strcat(current, s);
    cend = current + newlength;
    free(previous_buffer);
  }
  else {
    /* we have to copy the string out into a new buffer to insure
       it's properly null-terminated.  *&^%$#@!
    */
    current = newstring(len+1);
    strncpy(current, s, len);
    cend = current + len;
  }
  ctail = current;

  /* Loop trying to peel entries off the buffer.  Save the last one to
     tack onto the next buffer.  This does mean we get to do a special
     post-processing step for the last entry, but that's also forced
     on us by the funky CDATA callback semantics.
  */
  for(;;) {
    char *newtail = (*process_next)(ctail);
    if(newtail == ctail) {
      /* there was an error, report it */
      fprintf(stderr, "Value error near \"%30s\"...\n", ctail);
      PrintXMLLocation();
      exit(-1);
    }
    else if(newtail == cend) {
      /* current buffer exhausted - stash it */
      previous_buffer = current;
      previous_tail = ctail;
      break;
    }
    else {
      /* normal processing case */
      (*finish_processing)();
      ctail = newtail;
    }
  }
  /* I guess we're done... */
}

void post_process_in_line_buffer(void)
{
  /* At this point, inline_dataindex should be either 1 (process the last
     number) or 0 (there was whitespace at the end of the buffer, so we
     just punt)
  */
  switch(inline_datacount) {
  case 0: break;
  case 1:
    if((*process_next)(previous_tail) != previous_tail)
      (*finish_processing)();
    else {
      /* there was an error, report it */
      fprintf(stderr, "Value error near \"%30s\"...\n", previous_tail);
      PrintXMLLocation();
      exit(-1);
    }
    break;
  default:
    fprintf(stderr, "Ran out of in_line_data (needed %ld more values)\n",
	    inline_datacount);
      PrintXMLLocation();
      exit(-1);
      break;
  }
  /* regardless, turn off the CDATA handler */
  XML_SetCharacterDataHandler(TheParser, NULL);
  /* and free previous_buffer if there's anything in it */
  if(previous_buffer != NULL)
    free(previous_buffer);
}
