/* minimal support for expat */

#include "expat.h"

/* a structure for collecting CDATA characters */

typedef struct cdatabuf {
  long size;
  long current;
  char *buffer;
} cdatabuf;

/* stuff to handle stacking of element handlers and element local data */

/* the API for element data is a little weird, due to the generalized
   brain-damage that is SAX in general and expat in particular.  In short:
   * You may SetElementData() right after you PushEndHandler(), and the
     pushed data will be returned by PopHandler() in the end handler.
   * If you GetElementData() in an end handler after PopHandler(), you
     get the element data for the ENCLOSING element.
*/

typedef struct handlerstack {
  struct handlerstack *next;
  XML_EndElementHandler eh;
  void *elementdata;
  const char *name;
} handlerstack;

handlerstack *new_hs(void);
void recycle_hs(handlerstack *foo);
int hs_length(void);
void push_end_handler(XML_Parser p, XML_EndElementHandler eh, const char *ehname);
void *pop_end_handler(XML_Parser p);
void set_end_handler(XML_Parser p);
void show_end_handler_stack(void);
void set_element_data(void *data);
void *get_element_data(void);
void collect_cdata(void *userData, const XML_Char *s, int len);
void start_collecting_cdata(XML_Parser p, cdatabuf *cdb);
void stop_collecting_cdata(XML_Parser p);
XML_Char const *find_attribute(const XML_Char **atts, char const *aname);

/* Function stubs to enable in_line_data processing */

void inline_data_init(void *buf, long count, hid_t native_type );
void process_in_line_buffer(void *userData, const XML_Char *s, int len);
void post_process_in_line_buffer(void);

/* #definitions to hide a bunch of otherwise ugly names */

#define FindAttribute(x) find_attribute(atts, x)

#define PrintXMLLocation() fprintf(stderr, "XML parse error at line %d, column %d:\n", XML_GetCurrentLineNumber(TheParser), XML_GetCurrentColumnNumber(TheParser))

/* Macros to handle boiler-plate start-callback and end-callback */

#define SEHname(x) SEH_ ## x
#define SEHdeclare(x) void SEHname(x) (void *userData, const XML_Char *name, const XML_Char **atts)

#define EEHname(x) EEH_ ## x
#define EEHdeclare(x) void EEHname(x) (void *userData, const XML_Char *name)

#define CDHname(x) CDH_ ## x
#define CDHdeclare(x) void CDHname(x)(void *userData, const XML_Char *s, int len);


/* It doesn't hurt to declare them both, even if the end handler doesn't
   really exist */
#define HandlerDeclare(x) SEHdeclare(x); EEHdeclare(x)

#define SetStartHandler(x) set_start_handler(SEHname(x))
#define set_start_handler(x) XML_SetStartElementHandler(TheParser, x)
#define SetEndHandler(x) XML_SetEndElementHandler(TheParser, EEHname(x))
#define PushEndHandler(x) push_end_handler(TheParser, EEHname(x), "" # x)
#define PopHandler() pop_end_handler(TheParser)
#define TryHandler(x) SEHname(x)(userData, name, atts)
#define SetCDataHandler(x) XML_SetCharacterDataHandler(TheParser, CDHname(x))
#define SetElementData(x) set_element_data(x)
#define GetElementData() get_element_data()

/* A couple of common end-element handlers */
#define EEHpop(x) EEHdeclare(x) { TRACE((stderr, "Leaving " #x " with name %s\n", name)); PopHandler(); }
#define EEHnext(x, y) EEHdeclare(x) { TRACE((stderr, "Leaving " #x " for " #y " with name %s\n", name)); PopHandler(); SetStartHandler(y); }

#define NameIs(x) (!strcmp(name, x))
#define CheckName(x) if(!NameIs(x)) {fprintf(stderr, "Expected element %s, saw %s\n", x, name); PrintXMLLocation(); exit(-1);}



/* and some macros for using it */
#define StartCollectingCDATA() start_collecting_cdata(TheParser, &GlobalCDATA)
#define StopCollectingCDATA() stop_collecting_cdata(TheParser)

#define StartCollectingInLineData(a,b,c) inline_data_init(a,b,c)
#define StopCollectingInLineData() post_process_in_line_buffer()
