/* HDF-EOS support for the XML->HDF-EOS converter */
#include "hdfeos-support.h"

#include "support.h"

/* this is the inverse of the type_info mapping used in the validator */

hid_t hdfeos_native_type(H5T_class_t class_type, int size, int signp)
{
  hid_t result;

  /* Interesting stuff: I had to look at the HDF source code to determine
     that 8-bits == SCHAR/UCHAR, 16-bits == SHORT/USHORT,
     32-bits == INT/UNIT, but 64-bits == LLONG/ULLONG
     (32-bits is also == LONG/ULONG)
  */

  switch(class_type) {
  case H5T_INTEGER:
    switch (size) {
    case 1: result = (signp ?  H5T_NATIVE_SCHAR :H5T_NATIVE_UCHAR);
      break;
    case 2:result = (signp ?  H5T_NATIVE_SHORT :H5T_NATIVE_USHORT);
      break;
    case 4:result = (signp ?  H5T_NATIVE_INT :H5T_NATIVE_UINT);
      break;
    case 8:result = (signp ?  H5T_NATIVE_LLONG :H5T_NATIVE_ULLONG);
      break;
    default:
      fprintf(stderr, "Not an integer size: %d\n", size);
      exit(-1);
    }
    break;
  case H5T_FLOAT:
    switch(size) {
    case 4: result = H5T_NATIVE_FLOAT;
      break;
    case 8: result = H5T_NATIVE_DOUBLE;
      break;
      /* again from the HDF source - apparently 80-bits == LDOUBLE
         I've never seen this format anywhere (I believe it's Intel
	 internal) and think it's safe to ignore it in portable code.
      */
    default:
      fprintf(stderr, "Not a floating-point size: %d\n", size);
      exit(-1);
    }
    break;
  case H5T_STRING:
    result = H5T_NATIVE_CHAR;
    break;
  default:
    fprintf(stderr, "H5T_class_t type not handled: %d\n", class_type);
    exit(-1);
  }

  TRACE((stderr, "hdfeos_native_type for %d/size %d/sign %d is %d\n",
	 class_type, size, signp, result));
  return result;
}
    
/* support for dimension maps */

void add_dimension (dmap **dm, char const *name, hsize_t size)
{
  dmap *newd = calloc(1, sizeof(dmap));

  newd->name = strclone(name);
  newd->size = size;
  newd->next = *dm;
  *dm = newd;
}

/* Returns size of dimension (or 0 if not found, on the rash assumption
   that real zero-sized dimensions will be rare in reality */

hsize_t find_dimension(dmap *dm, char const *name)
{
  for( ; dm != NULL; dm = dm->next)
    if(!strcmp(name, dm->name))
      return dm->size;
  return 0;
}

void recycle_dm(dmap **dm)
{
  while(*dm != NULL) {
    dmap *dml = *dm;
    free_safely(dml->name);
    *dm = (*dm)->next;
    free(dml);
  }
}
    
