/* generic C support */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* Verbose and somewhat useful malloc/free tracing */
/* turn it on by hand-munging the obvious spot in headers.h */

#ifdef TRACE_MALLOC
void *MYmalloc(size_t size, int line, char *file) {
  void *result = malloc(size);
  fprintf(stderr, "Allocated %6d from %x (at %d in %s)\n", size, result, line, file);
  return result;
}
char *MYcalloc(size_t nmemb, size_t size, int line, char *file) {
  char *result = calloc(nmemb,size);
  fprintf(stderr, "Allocated %6d from %x (at %d in %s)\n", nmemb*size, result, line, file);
  return result;
}
void MYfree(void *ptr, int line, char *file){
  fprintf(stderr, "freeing %x (at %d in %s)\n", ptr, line, file);
  free(ptr);
}
#endif

#include "support.h"

/* code support for SS error "handling" */
void spit_status(int stat, char const *call) {
  if(stat < 0) {
    fprintf(stderr, "ERROR: code %d result of %s\n", stat, call);
    exit(-1);
  }
  fprintf(stderr, "RAN %s\n", call);
}

/* We use calloc to allocate string buffers because many of the HDF-EOS
   string-buffer-smashing routines do *NOT* null-terminate the strings
   they smash into the buffers.  I can see why not (FORTRAN compatibility),
   but it does make for counter-intuitive C processing...
*/
char *newstring(int len) {
  char *result = calloc((unsigned)len, 1);
  if(result == NULL)
    spit_status(-1, "Ran out of memory allocating a little bitty string!");
  return result;
}

/* a little string copier - surely this is a library function somewhere? */
char *strclone(const char *in)
{
  char *result = (char *) malloc(strlen(in) + 1);
  strcpy(result, in);
  return result;
}

/* "Safe" freeing of memory */
void free_safely(void *x) {
  if(x != NULL) free(x);
}
