/* Handles the callbacks for <Attribute> and its children */

#include "xml2he5.h"
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include "globals.h"

/* Handler stub that insures FileAtribute end handler gets called */
SEHdeclare(FileAttribute)
{
  TRACE((stderr, "Entering FileAttribute\n"));
  if(NameIs("Attribute")) {
    PushEndHandler(FileAttribute);
    TryHandler(Attribute);
  }
  else
    TryHandler(POINTS);
}

EEHdeclare(FileAttribute)
{

  PopHandler();
  TRACE((stderr, "Exiting FileAttribute\n"));

  /* Snarf/assert the attribute here */
  TRACE((stderr, "Setting file attribute %s (type %d) (count %ld)\n",
	 attribute_name, datatype_native_type, (long int)datatype_count));
  SS(HE5_EHwriteglbattr(hdfeos_handle,
			attribute_name, datatype_native_type,
			&datatype_count, attribute_buffer));
  free_safely(attribute_name);
  free_safely(attribute_buffer);

  /* Handling the loop */
  SetStartHandler(FileAttribute);
}

/* Handler stub that insures SwathAtribute end handler gets called */
SEHdeclare(SwathAttribute)
{
  TRACE((stderr, "Entering SwathAttribute\n"));
  if(NameIs("Attribute")) {
    PushEndHandler(SwathAttribute);
    TryHandler(Attribute);
  }
  else
    TryHandler(SwathDimension);
}

EEHdeclare(SwathAttribute)
{

  PopHandler();
  TRACE((stderr, "Exiting SwathAttribute\n"));

  /* Snarf/assert the attribute here */
  TRACE((stderr, "Setting file attribute %s (type %d) (count %ld)\n",
	 attribute_name, datatype_native_type, (long int)datatype_count));
  SS(HE5_SWwriteattr(swath_handle,
			attribute_name, datatype_native_type,
			&datatype_count, attribute_buffer));
  free_safely(attribute_name);
  free_safely(attribute_buffer);

  /* Handling the loop */
  SetStartHandler(SwathAttribute);
}


/* variable used by FieldAttribute end handler to control looping */
void (*field_loop_handler)(void *userData, const XML_Char *name, const XML_Char **atts);

/* Handler stub that insures GeoFieldAtribute end handler gets called */
SEHdeclare(GeoFieldAttribute)
{
  TRACE((stderr, "Entering GeoFieldAttribute\n"));
  if(NameIs("Attribute")) {
    PushEndHandler(FieldAttribute);
    field_loop_handler = SEHname(GeoFieldAttribute);
    TryHandler(Attribute);
  }
  else
    TryHandler(GeoFieldDataType);
}

/* Handler stub that insures DataFieldAtribute end handler gets called */
SEHdeclare(DataFieldAttribute)
{
  TRACE((stderr, "Entering DataFieldAttribute\n"));
  if(NameIs("Attribute")) {
    PushEndHandler(FieldAttribute);
    field_loop_handler = SEHname(DataFieldAttribute);
    TryHandler(Attribute);
  }
  else
    TryHandler(DataFieldDataType);
}

/* Handler stub that insures GridAtribute end handler gets called */
SEHdeclare(GridAttribute)
{
  TRACE((stderr, "Entering GridAttribute\n"));
  if(NameIs("Attribute")) {
    PushEndHandler(FieldAttribute);
    field_loop_handler = SEHname(GridAttribute);
    TryHandler(Attribute);
  }
  else
    TryHandler(Parameters);
}

/* Handler stub that insures GridDataAtribute end handler gets called */
SEHdeclare(GridDataAttribute)
{
  TRACE((stderr, "Entering GridDataAttribute\n"));
  if(NameIs("Attribute")) {
    PushEndHandler(FieldAttribute);
    field_loop_handler = SEHname(GridDataAttribute);
    TryHandler(Attribute);
  }
  else
    TryHandler(GridDataDataType);
}

/* This gets used by Attributes in both GeoField and DataField elements */
EEHdeclare(FieldAttribute)
{
  attributestack *as = (attributestack *)malloc(sizeof(attributestack));

  PopHandler();
  TRACE((stderr, "Exiting GeoFieldAttribute\n"));

  /* Snarf/assert the attribute here */
  as->name = attribute_name;
  as->type = datatype_native_type;
  as->count = datatype_count;
  as->buffer = attribute_buffer;

  as->next = field_attributes;
  field_attributes = as;

  /* Handling the loop */
  set_start_handler(field_loop_handler);
}

/* Generic Attribute handler - accumulates current attribute type/size/value
   in global variables, and end handler grabs and asserts them
   <!ELEMENT Attribute (DataType, Count, Value+)>
*/
SEHdeclare(Attribute)
{
  XML_Char const *attname;

  TRACE((stderr, "Entering Attribute\n"));
  CheckName("Attribute");
  /* Whoever sent us here has set our end handler already */
  /* snarf the attribute name */
  if ((attname = FindAttribute("Name")) == NULL) {
    fprintf(stderr, "Attribute with no Name attribute\n");
    PrintXMLLocation();
    exit(-1);
  }
  attribute_name = strclone(attname);
  SetStartHandler(AttributeDataType);
}

/* <!ELEMENT DataType (IntegerType | FloatType | StringType | TimeType | BitfieldType | OpaqueType | ReferenceType | EnumType)> */
SEHdeclare(AttributeDataType)
{
  TRACE((stderr, "Entering Attribute DataType\n"));
  CheckName("DataType");
  PushEndHandler(AttributeDataType);
  SetStartHandler(IntegerType);
}


EEHdeclare(AttributeDataType)
{
  PopHandler();
  SetStartHandler(Count);
}

SEHdeclare(GeoFieldDataType)
{
  TRACE((stderr, "Entering GeoFieldDataType\n"));
  CheckName("DataType");
  PushEndHandler(GeoFieldDataType);
  SetStartHandler(IntegerType);
}


EEHdeclare(GeoFieldDataType)
{
  PopHandler();
  SetStartHandler(GeoFieldDimList);
}

SEHdeclare(DataFieldDataType)
{
  TRACE((stderr, "Entering DataFieldDataType\n"));
  CheckName("DataType");
  PushEndHandler(DataFieldDataType);
  SetStartHandler(IntegerType);
}


EEHdeclare(DataFieldDataType)
{
  PopHandler();
  SetStartHandler(DataFieldDimList);
}

SEHdeclare(GridDataDataType)
{
  TRACE((stderr, "Entering GridDataDataType\n"));
  CheckName("DataType");
  PushEndHandler(GridDataDataType);
  SetStartHandler(IntegerType);
}


EEHdeclare(GridDataDataType)
{
  PopHandler();
  SetStartHandler(GridDataDimList);
}


/* <!ELEMENT Count (#PCDATA)> */
SEHdeclare(Count)
{
  TRACE((stderr, "Entering Count\n"));
  CheckName("Count");
  PushEndHandler(Count);
  StartCollectingCDATA();
}

EEHdeclare(Count)
{
  PopHandler();
  StopCollectingCDATA();
  /* scan off the count */
  if(sscanf(GlobalCDATA.buffer, "%ld", &datatype_count) != 1) {
    fprintf(stderr, "Not a number for Count: %s\n", GlobalCDATA.buffer);
    PrintXMLLocation();
    exit(-1);
  }
  /* And get ready to gobble up the Values */
  attribute_buffer = (void *)malloc(datatype_count * datatype_size + 1);
  attribute_ptr = attribute_buffer;
  switch(datatype_type) {
  case H5T_INTEGER:
    SetStartHandler(intValue);
    break;
  case H5T_FLOAT:
    SetStartHandler(floatValue);
    break;
  case H5T_STRING:
    SetStartHandler(stringValue);
    break;
  default:
    fprintf(stderr, "Can't handle attribute of type %d; sorry!\n", datatype_type);
    exit(-1);
  }
}

/* Stub that insures Value gets interpreted as integer */
SEHdeclare(intValue)
{
  TRACE((stderr, "Entering intValue\n"));
  CheckName("Value");
  PushEndHandler(intValue);
  StartCollectingCDATA();
}

EEHdeclare(intValue)
{
  long ival;

  PopHandler();
  StopCollectingCDATA();
  TRACE((stderr, "Exiting intValue to parse %s\n", GlobalCDATA.buffer));
 
  if(sscanf(GlobalCDATA.buffer, "%ld", &ival) != 1) {
    fprintf(stderr, "Not a number: %s\n", GlobalCDATA.buffer);
    PrintXMLLocation();
    exit(-1);
  }

  switch(datatype_size) {
  case 1: *((char *)attribute_ptr)++ = ival;
    break;
  case 2: *((short *)attribute_ptr)++ = ival;
    break;
  case 4: *((int *)attribute_ptr)++ = ival;
    break;
  case 8: *((long *)attribute_ptr)++ = ival;
    break;
  default:
    fprintf(stderr, "Integer length %d not supported, sorry!\n", datatype_size);
    exit(-1);
  }
  
  SetStartHandler(intValue);
}

/* Stub that insures Value gets interpreted as float */
SEHdeclare(floatValue)
{
  CheckName("Value");
  PushEndHandler(floatValue);
  StartCollectingCDATA();
}

EEHdeclare(floatValue)
{
  double dval;

  PopHandler();
  StopCollectingCDATA();
 
  if(sscanf(GlobalCDATA.buffer, "%lf", &dval) != 1) {
    fprintf(stderr, "Not a number: %s\n", GlobalCDATA.buffer);
    PrintXMLLocation();
    exit(-1);
  }

  switch(datatype_size) {
  case 4: *((float *)attribute_ptr)++ = dval;
    break;
  case 8: *((double *)attribute_ptr)++ = dval;
    break;
  default:
    fprintf(stderr, "Float length %d not supported, sorry!\n", datatype_size);
    exit(-1);
  }
  
  SetStartHandler(floatValue);
}

/* Stub that insures Value gets interpreted as string */
SEHdeclare(stringValue)
{
  CheckName("Value");
  PushEndHandler(stringValue);
  StartCollectingCDATA();
}

EEHdeclare(stringValue)
{
  PopHandler();
  StopCollectingCDATA();
 
  ((char*)attribute_buffer)[0] = 0;
  strncat(attribute_buffer, GlobalCDATA.buffer, datatype_count);
}

/* <!ELEMENT IntegerType EMPTY> */
SEHdeclare(IntegerType)
{
  if(NameIs("IntegerType")) {
    const XML_Char **att;
    int signp = -1;

    datatype_size = -1;
    PushEndHandler(JustPop);
    TRACE((stderr, "Entering IntegerType\n"));

    datatype_type = H5T_INTEGER;
    for(att = atts; *att != NULL; att = att + 2) {
      if(!strcmp(*att, "Size")) {
	if(sscanf(*(att+1), "%d", &datatype_size) != 1
	   || !(datatype_size == 1 || datatype_size == 2
		|| datatype_size == 4 || datatype_size == 8)) {
	  fprintf(stderr, "Not an integer size: %s\n", (*att+1));
	  PrintXMLLocation();
	  exit(-1);
	}
      }
      else if(!strcmp(*att, "Sign")) {
	if(!strcmp(*(att+1), "true"))
	  signp = 1;
	else if(!strcmp(*(att+1), "false"))
	  signp = 0;
	else {
	  fprintf(stderr, "Not true or false: %s\n", *(att+1));
	  PrintXMLLocation();
	  exit(-1);
	}
      }
      else if(!strcmp(*att, "ByteOrder")) {
	if(strcmp(*(att+1), "BE") && strcmp(*(att+1), "LE")) {
	  fprintf(stderr, "Not a byte order: %s\n", *(att+1));
	  PrintXMLLocation();
	  exit(-1);
	}
      }
      else {
	fprintf(stderr, "Not an IntegerType attribute: %s\n", *att);
	PrintXMLLocation();
	exit(-1);
      }
    }
    if(signp == -1) {
      fprintf(stderr, "Integer type with no Sign attribute\n");
      PrintXMLLocation();
      exit(-1);
    }
    if(datatype_size == -1) {
      fprintf(stderr, "IntegerType with no Size attribute\n");
      PrintXMLLocation();
      exit(-1);
    }
    datatype_native_type = hdfeos_native_type(datatype_type, datatype_size, signp);
  }
  else
    TryHandler(FloatType);
}

/* <!ELEMENT FloatType EMPTY> */
SEHdeclare(FloatType)
{
  if(NameIs("FloatType")) {
    const XML_Char **att;

    PushEndHandler(JustPop);

    datatype_type = H5T_FLOAT;
    datatype_size = -1;

    for(att = atts; *att != NULL; att = att + 2) {
      if(!strcmp(*att, "Size")) {
	if(sscanf(*(att+1), "%d", &datatype_size) != 1
	   || !(datatype_size == 4 || datatype_size == 8)) {
	  fprintf(stderr, "Not a floating-point size: %s\n", (*att+1));
	  PrintXMLLocation();
	  exit(-1);
	}
      }
      else if(strcmp((*att), "ByteOrder")
	      && strcmp((*att), "SignBitLocation")
	      && strcmp((*att), "ExponentBits")
	      && strcmp((*att), "ExponentLocation")
	      && strcmp((*att), "MantissaBits")
	      && strcmp((*att), "TypeCode")) {
	fprintf(stderr, "Not a FloatType attribute: %s\n", *att);
	PrintXMLLocation();
	exit(-1);
      }
    }
    if(datatype_size == -1) {
      fprintf(stderr, "FloatType with no Size attribute\n");
      PrintXMLLocation();
      exit(-1);
    }
    datatype_native_type = hdfeos_native_type(datatype_type, datatype_size, 1);
	
  }
  else
    TryHandler(StringType);
}

/* <!ELEMENT StringType EMPTY> */
SEHdeclare(StringType)
{
  if(NameIs("StringType")) {
    const XML_Char **att;

    PushEndHandler(JustPop);

    datatype_type = H5T_STRING;
    datatype_count = (unsigned)-1;

    for(att = atts; *att != NULL; att = att + 2) {
      if(!strcmp(*att, "StrSize")) {
	if(sscanf(*(att+1), "%ld", &datatype_count) != 1) {
	  fprintf(stderr, "Not a string length: %s\n", *(att+1));
	  PrintXMLLocation();
	  exit(-1);
	}
      }
      else if(strcmp(*att, "Cset")
	      && strcmp(*att, "StrPad")) {
	fprintf(stderr, "Not a StringType attribute: %s\n", *att);
	PrintXMLLocation();
	exit(-1);
      }
    }
    if(datatype_count == (unsigned)-1) {
      fprintf(stderr, "StringType with no StrSize attribute\n");
      PrintXMLLocation();
      exit(-1);
    }
    datatype_native_type = H5T_NATIVE_CHAR;
    datatype_size = 1;
  }
  else
    TryHandler(OtherType);
}

/* Stub that punts TimeType | BitfieldType | OpaqueType | ReferenceType | EnumType */
SEHdeclare(OtherType)
{
  fprintf(stderr, "Can't currently handle attribute of %s type; sorry\n",
	  name);
  exit(-1);
}
