#include "xml2he5.h"
#include <stdio.h>

#include "globals.h"

/* <!ELEMENT FieldStorage (Merge?, Compression?)> */

SEHdeclare(FieldStorage) {

  CheckName("FieldStorage");
  SetStartHandler(Merge);
}

SEHdeclare(Merge)
{
  if(NameIs("Merge")) {
    XML_Char const *mergecode = FindAttribute("Code");

    if(mergecode == NULL) {
      fprintf(stderr, "Merge with no Code attribute\n");
      PrintXMLLocation();
      exit(-1);
    }
    if(!strcmp(mergecode, "HDFE_NOMERGE"))
      fieldstorage_mergecode = 0;
    else if(!strcmp(mergecode, "HDFE_AUTOMERGE"))
      fieldstorage_mergecode = 1;
    else {
      fprintf(stderr, "Merge with bogus Code attribute %s\n", mergecode);
      PrintXMLLocation();
      exit(-1);
    }
    PushEndHandler(JustPop);
    SetStartHandler(Compression);
  }
  else
    TryHandler(Compression);
}

int parmcount;

SEHdeclare(Compression)
{
  if(NameIs("Compression")) {
    XML_Char const *compcode = FindAttribute("Code");

    if(compcode == NULL) {
      fprintf(stderr, "Compression with no Code attribute\n");
      PrintXMLLocation();
      exit(-1);
    }
    if(!strcmp(compcode, "HDFE_COMP_NONE"))
      fieldstorage_compcode = 0;
    else if(!strcmp(compcode, "HDFE_COMP_RLE"))
      fieldstorage_compcode = 1;
    else if(!strcmp(compcode, "HDFE_COMP_NBIT"))
      fieldstorage_compcode = 2;
    else if(!strcmp(compcode, "HDFE_COMP_SKPHUFF"))
      fieldstorage_compcode = 3;
    else if(!strcmp(compcode, "HDFE_COMP_DEFLATE"))
      fieldstorage_compcode = 4;
    else {
      fprintf(stderr, "Compression with bogus Code attribute %s\n", compcode);
      PrintXMLLocation();
      exit(-1);
    }

    /* and get ready to vacuum up any parameters */
    fieldstorage_compparms = NULL;
    parmcount = 0;
    PushEndHandler(JustPop);
    SetStartHandler(CompParameter);
  }
  else{
    fprintf(stderr, "Bogus element %s in FieldStorage\n", name);
    PrintXMLLocation();
    exit(-1);
  }
}

SEHdeclare(CompParameter)
{
  TRACE((stderr, "Entering CompParameter\n"));
  CheckName("CompParameter");
  PushEndHandler(CompParameter);
  StartCollectingCDATA();
}

EEHdeclare(CompParameter)
{
  int cparam;

  PopHandler();
  StopCollectingCDATA();
  if(sscanf(GlobalCDATA.buffer, "%d", &cparam) != 1) {
    fprintf(stderr, "Not a parameter number: %s\n", GlobalCDATA.buffer);
    PrintXMLLocation();
    exit(-1);
  }

  /* extend the parameter array by 1 integer, smack the new value into it */
  fieldstorage_compparms = realloc(fieldstorage_compparms,
				   sizeof(int) * (++parmcount));
  fieldstorage_compparms[parmcount -1] = cparam;
}
