/* Handles the callbacks for <Grid> and its children */

#include "xml2he5.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "globals.h"

/* Pre-declarations for the functions in this file */

HandlerDeclare(GridDimension);
HandlerDeclare(YDim);
HandlerDeclare(XDim);
HandlerDeclare(UpperLeftPoint);
HandlerDeclare(LowerRightPoint);
HandlerDeclare(Projection);
HandlerDeclare(GEO);
HandlerDeclare(UTM);
HandlerDeclare(UtmParameters);
HandlerDeclare(StuffParameter);
HandlerDeclare(Longitude_Point);
HandlerDeclare(Latitude_Point);
HandlerDeclare(LAMCC);
HandlerDeclare(lamccParameters);
HandlerDeclare(lamccSemi_Major_Axis);
HandlerDeclare(Semi_Major_Axis);
HandlerDeclare(lamccSemi_Minor_Axis);
HandlerDeclare(Semi_Minor_Axis);
HandlerDeclare(Latitude_First_Standard_Parallel);
HandlerDeclare(Latitude_Second_Standard_Parallel);
HandlerDeclare(PS);
HandlerDeclare(psParameters);
HandlerDeclare(psSemi_Major_Axis);
HandlerDeclare(psSemi_Minor_Axis);
HandlerDeclare(Longitude_Below_Pole);
HandlerDeclare(Latitude_True_Scale);
HandlerDeclare(psFalse_Easting);
HandlerDeclare(False_Easting);
HandlerDeclare(psFalse_Northing);
HandlerDeclare(False_Northing);
HandlerDeclare(POLYC);
HandlerDeclare(polycParameters);
HandlerDeclare(polycSemi_Major_Axis);
HandlerDeclare(polycSemi_Minor_Axis);
HandlerDeclare(polycLongitude_Central_Meridian);
HandlerDeclare(Longitude_Central_Meridian);
HandlerDeclare(polycLatitude_Origin);
HandlerDeclare(Latitude_Origin);
HandlerDeclare(polycFalse_Easting);
HandlerDeclare(polycFalse_Northing);
HandlerDeclare(TM);
HandlerDeclare(tmParameters);
HandlerDeclare(tmSemi_Major_Axis);
HandlerDeclare(tmSemi_Minor_Axis);
HandlerDeclare(tmCentral_Meridian_Scale_Factor);
HandlerDeclare(Central_Meridian_Scale_Factor);
HandlerDeclare(tmLongitude_Central_Meridian);
HandlerDeclare(tmLatitude_Origin);
HandlerDeclare(tmFalse_Easting);
HandlerDeclare(tmFalse_Northing);
HandlerDeclare(LAMAZ);
HandlerDeclare(lamazParameters);
HandlerDeclare(lamazReference_Sphere_Radius);
HandlerDeclare(Reference_Sphere_Radius);
HandlerDeclare(Longitude_Center);
HandlerDeclare(Latitude_Center);
HandlerDeclare(lamazFalse_Easting);
HandlerDeclare(lamazFalse_Northing);
HandlerDeclare(HOM);
HandlerDeclare(hom_aParameters);
HandlerDeclare(hom_bParameters);
HandlerDeclare(hom_aSemi_Major_Axis);
HandlerDeclare(hom_aSemi_Minor_Axis);
HandlerDeclare(hom_aCentral_Meridian_Scale_Factor);
HandlerDeclare(hom_aLatitude_Origin);
HandlerDeclare(hom_aFalse_Easting);
HandlerDeclare(hom_aFalse_Northing);
HandlerDeclare(Longitude_Center_Line1);
HandlerDeclare(Longitude_Center_Line2);
HandlerDeclare(Latitude_Center_Line1);
HandlerDeclare(Latitude_Center_Line2);
HandlerDeclare(hom_aformat_A);
HandlerDeclare(format_A);
HandlerDeclare(hom_bSemi_Major_Axis);
HandlerDeclare(hom_bSemi_Minor_Axis);
HandlerDeclare(hom_bCentral_Meridian_Scale_Factor);
HandlerDeclare(Azimuth_Angle);
HandlerDeclare(Azimuth_Longitude);
HandlerDeclare(hom_bLatitude_Origin);
HandlerDeclare(hom_bFalse_Easting);
HandlerDeclare(hom_bFalse_Northing);
HandlerDeclare(hom_bformat_B);
HandlerDeclare(format_B);
HandlerDeclare(SOM);
HandlerDeclare(som_aParameters);
HandlerDeclare(som_bParameters);
HandlerDeclare(som_aSemi_Major_Axis);
HandlerDeclare(som_aSemi_Minor_Axis);
HandlerDeclare(Orbit_Inclination);
HandlerDeclare(Longitude_Ascending_Orbit);
HandlerDeclare(som_aFalse_Easting);
HandlerDeclare(som_aFalse_Northing);
HandlerDeclare(Period_of_Revolution);
HandlerDeclare(Satellite_Ratio);
HandlerDeclare(Landsat_End_Of_Path);
HandlerDeclare(som_aHDF_EOS_Variable);
HandlerDeclare(HDF_EOS_Variable);
HandlerDeclare(som_aformat_A);
HandlerDeclare(som_bSemi_Major_Axis);
HandlerDeclare(som_bSemi_Minor_Axis);
HandlerDeclare(Landsat_Satellite_Number);
HandlerDeclare(Landsat_Path_Number);
HandlerDeclare(som_bFalse_Easting);
HandlerDeclare(som_bFalse_Northing);
HandlerDeclare(som_bHDF_EOS_Variable);
HandlerDeclare(som_bformat_B);
HandlerDeclare(GOOD);
HandlerDeclare(goodParameters);
HandlerDeclare(goodReference_Sphere_Radius);
HandlerDeclare(ISINUS);
HandlerDeclare(isinusParameters);
HandlerDeclare(isinusReference_Sphere_Radius);
HandlerDeclare(isinusLongitude_Central_Meridian);
HandlerDeclare(isinusFalse_Easting);
HandlerDeclare(isinusFalse_Northing);
HandlerDeclare(Latitudinal_Zones);
HandlerDeclare(Right_Justified_Columns);
HandlerDeclare(GridDataField);
HandlerDeclare(TileDimList);
HandlerDeclare(TileDim);
HandlerDeclare(GridDatalink_to_data);
HandlerDeclare(GridDatain_line_data);
HandlerDeclare(SphereCode);
HandlerDeclare(ZoneCode);
HandlerDeclare(PixelRegistration);
HandlerDeclare(OriginType);
void check_one_attribute(const XML_Char **atts, char const *pname, char const *attname, char const *attval);
void check_proj_attributes(const XML_Char **atts, char const *pname, char const *pcode, char const *pfullname, char const *pid);
void grab_param_index(const XML_Char **atts, char const *pname);

/* and a little stuff outside the file */
void *inhale_file(void);

/* <!ELEMENT Grid (Attribute*, Parameters, Dimension*, GridDataField)> */
SEHdeclare(Grid)
{

  XML_Char const *xc_grid_name = FindAttribute("Name");

  TRACE((stderr, "Entering Grid\n"));
  CheckName("Grid");
  PushEndHandler(Grid);

  if(xc_grid_name == NULL) {
    fprintf(stderr, "Grid with no Name attribute\n");
    PrintXMLLocation();
    exit(-1);
  }
  grid_name = strclone(xc_grid_name);

  /* Set up to remember any dimensions that get defined */
  recycle_dm(&dimension_map);

  /* Set up to eat any Attributes */
  SetStartHandler(GridAttribute);
}

EEHdeclare(Grid)
{
  PopHandler();
  TRACE((stderr, "Exiting Grid\n"));

  /* Close the Grid */
  SS(HE5_GDdetach(grid_handle));
  /* clean up after ourselves */
  free_safely(grid_name);
  free_safely(griddimnames);
  free_safely(griddimsizes);
  free_safely(gridmaxdimnames);
  free_safely(gridmaxdimsizes);

  /* Handle the next Grid, if any */
  SetStartHandler(Grid);
}

/* variables to hold the Parameter data */
long ydim;
long xdim;
double upleft[2];
double lowright[2];
int projcode;
int zonecode;
int spherecode;
double projparm[13];
int prcode;
int origincode;

/* <!ELEMENT Parameters (YDim, XDim UpperLeftPoint, LowerRightPoint, Projection, SphereCode, ZoneCode, PixelRegistration, OriginType)> */
SEHdeclare(Parameters)
{
  TRACE((stderr, "Entering Parameters\n"));
  CheckName("Parameters");
  PushEndHandler(Parameters);
  SetStartHandler(YDim);
}

EEHdeclare(Parameters)
{
  PopHandler();
  TRACE((stderr, "Exiting Parameters\n"));
  /* A bunch of tedious creation of crap, now that we have everything */

  SS(grid_handle = HE5_GDcreate(hdfeos_handle, grid_name,
				xdim, ydim, upleft, lowright));
  SS(HE5_GDdeforigin(grid_handle, origincode));
  SS(HE5_GDdefpixreg(grid_handle, prcode));
  SS(HE5_GDdefproj(grid_handle, projcode, zonecode, spherecode, projparm));
  /* must manually create XDim and YDim entries in the dimension_map */
  add_dimension(&dimension_map, "XDim", xdim);
  add_dimension(&dimension_map, "YDim", ydim);

  /* Now that the grid is created, we can unroll its attributes, if any */
  while(field_attributes != NULL) {
    attributestack *as = field_attributes;
    SS(HE5_GDwriteattr(grid_handle, as->name, as->type,
		       &(as->count), as->buffer));
    free_safely(as->name);
    free_safely(as->buffer);
    field_attributes = field_attributes->next;
    free_safely(as);
  }

  SetStartHandler(GridDimension);
}

SEHdeclare(YDim)
{
  TRACE((stderr, "Entering YDim\n"));
  CheckName("YDim");
  PushEndHandler(YDim);
  StartCollectingCDATA();
}

EEHdeclare(YDim)
{
  PopHandler();
  StopCollectingCDATA();
  if(sscanf(GlobalCDATA.buffer, "%ld", &ydim) != 1) {
    fprintf(stderr, "YDim not a number: %s\n", GlobalCDATA.buffer);
    PrintXMLLocation();
    exit(-1);
  }
  SetStartHandler(XDim);
}

SEHdeclare(XDim)
{
  TRACE((stderr, "Entering XDim\n"));
  CheckName("XDim");
  PushEndHandler(XDim);
  StartCollectingCDATA();
}

EEHdeclare(XDim)
{
  PopHandler();
  StopCollectingCDATA();
  if(sscanf(GlobalCDATA.buffer, "%ld", &xdim) != 1) {
    fprintf(stderr, "XDim not a number: %s\n", GlobalCDATA.buffer);
    PrintXMLLocation();
    exit(-1);
  }
  SetStartHandler(UpperLeftPoint);
}

SEHdeclare(UpperLeftPoint)
{
  XML_Char const *Xstring = FindAttribute("X");
  XML_Char const *Ystring = FindAttribute("Y");
  TRACE((stderr, "Entering UpperLeftPoint\n"));
  PushEndHandler(JustPop);
  if(Xstring == NULL) {
    fprintf(stderr, "Missing X attribute in UpperLeftPoint\n");
    PrintXMLLocation();
    exit(-1);
  }
  if(Ystring == NULL) {
    fprintf(stderr, "Missing X attribute in UpperLeftPoint\n");
    PrintXMLLocation();
    exit(-1);
  }
  if(sscanf(Xstring, "%lf", &upleft[0]) != 1) {
    fprintf(stderr, "UpperLeftPoint X not a number: %s\n", Xstring);
    PrintXMLLocation();
    exit(-1);
  }
  if(sscanf(Ystring, "%lf", &upleft[1]) != 1) {
    fprintf(stderr, "UpperLeftPoint Y not a number: %s\n", Ystring);
    PrintXMLLocation();
    exit(-1);
  }

  SetStartHandler(LowerRightPoint);
}

SEHdeclare(LowerRightPoint)
{
  XML_Char const *Xstring = FindAttribute("X");
  XML_Char const *Ystring = FindAttribute("Y");

  TRACE((stderr, "Entering LowerRightPoint\n"));
  PushEndHandler(JustPop);
  if(Xstring == NULL) {
    fprintf(stderr, "Missing X attribute in LowerRightPoint\n");
    PrintXMLLocation();
    exit(-1);
  }
  if(Ystring == NULL) {
    fprintf(stderr, "Missing X attribute in LowerRightPoint\n");
    PrintXMLLocation();
    exit(-1);
  }
  if(sscanf(Xstring, "%lf", &upleft[0]) != 1) {
    fprintf(stderr, "LowerRightPoint X not a number: %s\n", Xstring);
    PrintXMLLocation();
    exit(-1);
  }
  if(sscanf(Ystring, "%lf", &upleft[1]) != 1) {
    fprintf(stderr, "LowerRightPoint Y not a number: %s\n", Ystring);
    PrintXMLLocation();
    exit(-1);
  }

  SetStartHandler(Projection);
}
/* <!ELEMENT Projection (GEO | UTM | SPCS | ALBERS | LAMCC | MERCAT | PS | POLYC | EQUIDC | TM | STEREO | LAMAZ | AZMEQD | GNOMON | ORTHO | GVNSP | SNSOID | EQRECT | MILLER | VGRINT | HOM | ROBIN | SOM | ALASKA | GOOD | MOLL | IMOLL | HAMMER | WAGIV | WAGVII | OBLEQA | ISINUS)> */

/* Actually we only implement the ones declared legal in the HDF-EOS5
   documentation:
   GEO, UTM, LAMCC, PS, POLYC, TM, LAMAZ, HOM, SOM, GOOD, ISINUS
*/
SEHdeclare(Projection)
{
  int i;

  TRACE((stderr, "Entering Projection\n"));
  CheckName("Projection");
  PushEndHandler(Projection);
  SetStartHandler(GEO);
  for(i=0; i<13; i++)
    projparm[i] = 0.0;
  projcode = -1;
}

void check_one_attribute(const XML_Char **atts, char const *pname, char const *attname, char const *attval)
{
  XML_Char const *aval = FindAttribute(attname);

  if(aval == NULL) {
    fprintf(stderr, "Element %s missing attribute %s\n", pname, attname);
    PrintXMLLocation();
    exit(-1);
  }
  if(strcmp(aval, attval)) {
    fprintf(stderr, "Element %s attribute %s is %s, not required %s\n",
	    pname, attname, aval, attval);
    PrintXMLLocation();
    exit(-1);
  }
}

void check_proj_attributes(const XML_Char **atts, char const *pname, char const *pcode, char const *pfullname, char const *pid)
{

  check_one_attribute(atts, pname, "Code", pcode);
  check_one_attribute(atts, pname, "Name", pfullname);
  check_one_attribute(atts, pname, "ID", pid);
}

#define CheckProjAtts(n,c,f,i) check_proj_attributes(atts,n,c,f,i)
SEHdeclare(GEO)
{
  TRACE((stderr, "Entering GEO\n"));
  if(NameIs("GEO")) {
    PushEndHandler(JustPop);
    CheckProjAtts("GEO", "HE5_GCTP_GEO", "Geographic", "0");
    projcode = 0;
  }
  else TryHandler(UTM);
}

SEHdeclare(UTM)
{
  TRACE((stderr, "Entering UTM\n"));
  if(NameIs("UTM")) {
    PushEndHandler(JustPop);
    CheckProjAtts("UTM", "HE5_GCTP_UTM", "Universal Transverse Mercator (UTM)", "1");
    projcode = 1;
    SetStartHandler(UtmParameters);
  }
  else TryHandler(LAMCC);
}

SEHdeclare(UtmParameters)
{
  TRACE((stderr, "Entering UtmParameters\n"));
  CheckName("UtmParameters");
  PushEndHandler(JustPop);
  SetStartHandler(Longitude_Point);
}

int param_index;
char const *param_element_name;
void (*next_parameter_handler)(void *userData, const XML_Char *name, const XML_Char **atts);

void grab_param_index(const XML_Char **atts, char const *pname)
{
  const XML_Char *pval = FindAttribute("Index");

  param_element_name = pname;

  if(pval == NULL) {
    fprintf(stderr, "Element %s has no Index attribute\n", pname);
    PrintXMLLocation();
    exit(-1);
  }
  if(sscanf(pval, "%d", &param_index) != 1) {
    fprintf(stderr, "Element %s Index attribute is not a number: %s\n",
	    pname, pval);
    PrintXMLLocation();
    exit(-1);
  }
}

#define GrabParamIndex(n) grab_param_index(atts, n)

EEHdeclare(StuffParameter)
{
  PopHandler();
  StopCollectingCDATA();
  if(sscanf(GlobalCDATA.buffer, "%lf", &projparm[param_index]) != 1) {
    fprintf(stderr, "Parameter from element %s is not a number:%s",
	    param_element_name, GlobalCDATA.buffer);
    PrintXMLLocation();
    exit(-1);
  }
  set_start_handler(next_parameter_handler);
}

SEHdeclare(Longitude_Point)
{
  TRACE((stderr, "Entering Longitude_Point\n"));
  CheckName("Longitude_Point");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Longitude_Point");
  next_parameter_handler = SEHname(Latitude_Point);
}

SEHdeclare(Latitude_Point)
{
  TRACE((stderr, "Entering Latitude_Point\n"));
  CheckName("Latitude_Point");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Latitude_Point");
  next_parameter_handler = SEHname(NoMore);
}

SEHdeclare(LAMCC)
{
  TRACE((stderr, "Entering LAMCC\n"));
  if(NameIs("LAMCC")) {
    PushEndHandler(JustPop);
    CheckProjAtts("LAMCC", "HE5_GCTP_LAMCC", "Lambert Conformal Conic", "4");
    projcode = 4;
    SetStartHandler(lamccParameters);
  }
  else TryHandler(PS);
}

SEHdeclare(lamccParameters)
{
  TRACE((stderr, "Entering lamccParameters\n"));
  CheckName("lamccParameters");
  PushEndHandler(JustPop);
  SetStartHandler(lamccSemi_Major_Axis);
}

SEHdeclare(lamccSemi_Major_Axis)
{
  next_parameter_handler = SEHname(lamccSemi_Minor_Axis);
  TryHandler(Semi_Major_Axis);
}


SEHdeclare(Semi_Major_Axis)
{
  TRACE((stderr, "Entering Semi_Major_Axis\n"));
  CheckName("Semi_Major_Axis");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Semi_Major_Axis");
}

SEHdeclare(lamccSemi_Minor_Axis)
{
  next_parameter_handler = SEHname(Latitude_First_Standard_Parallel);
  TryHandler(Semi_Major_Axis);
}

SEHdeclare(Semi_Minor_Axis)
{
  TRACE((stderr, "Entering Semi_Minor_Axis\n"));
  CheckName("Semi_Minor_Axis");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Semi_Minor_Axis");
}

SEHdeclare(Latitude_First_Standard_Parallel)
{
  TRACE((stderr, "Entering Latitude_First_Standard_Parallel\n"));
  CheckName("Latitude_First_Standard_Parallel");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Latitude_First_Standard_Parallel");
  next_parameter_handler = SEHname(Latitude_First_Standard_Parallel);
}

SEHdeclare(Latitude_Second_Standard_Parallel)
{
  TRACE((stderr, "Entering Latitude_Second_Standard_Parallel\n"));
  CheckName("Latitude_Second_Standard_Parallel");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Latitude_Second_Standard_Parallel");
  next_parameter_handler = SEHname(NoMore);
}

SEHdeclare(PS)
{
  TRACE((stderr, "Entering PS\n"));
  if(NameIs("PS")) {
    PushEndHandler(JustPop);
    CheckProjAtts("PS", "HE5_GCTP_PS", "PolarStereographic", "6");
    projcode = 6;
    SetStartHandler(psParameters);
  }
  else TryHandler(POLYC);
}

SEHdeclare(psParameters)
{
  TRACE((stderr, "Entering psParameters\n"));
  CheckName("lamccParameters");
  PushEndHandler(JustPop);
  SetStartHandler(psSemi_Major_Axis);
}

SEHdeclare(psSemi_Major_Axis)
{
  next_parameter_handler = SEHname(psSemi_Minor_Axis);
  TryHandler(Semi_Major_Axis);
}

SEHdeclare(psSemi_Minor_Axis)
{
  next_parameter_handler = SEHname(Longitude_Below_Pole);
  TryHandler(Semi_Minor_Axis);
}

SEHdeclare(Longitude_Below_Pole)
{
  TRACE((stderr, "Entering Longitude_Below_Pole\n"));
  CheckName("Longitude_Below_Pole");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Longitude_Below_Pole");
  next_parameter_handler = SEHname(Latitude_True_Scale);
}

SEHdeclare(Latitude_True_Scale)
{
  TRACE((stderr, "Entering Latitude_True_Scale\n"));
  CheckName("Latitude_True_Scale");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Latitude_True_Scale");
  next_parameter_handler = SEHname(psFalse_Easting);
}

SEHdeclare(psFalse_Easting)
{
  next_parameter_handler = SEHname(psFalse_Northing);
  TryHandler(False_Easting);
}

SEHdeclare(False_Easting)
{
  TRACE((stderr, "Entering False_Easting\n"));
  CheckName("False_Easting");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("False_Easting");
}

SEHdeclare(psFalse_Northing)
{
  next_parameter_handler = SEHname(NoMore);
  TryHandler(False_Northing);
}

SEHdeclare(False_Northing)
{
  TRACE((stderr, "Entering False_Northing\n"));
  CheckName("False_Northing");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("False_Northing");
}

SEHdeclare(POLYC)
{
  TRACE((stderr, "Entering POLYC\n"));
  if(NameIs("POLYC")) {
    PushEndHandler(JustPop);
    check_proj_attributes(atts, "POLYC", "HE5_GCTP_POLYC", "Polyconic", "7");
    projcode = 7;
    SetStartHandler(polycParameters);
  }
  else TryHandler(TM);
}

SEHdeclare(polycParameters)
{
  TRACE((stderr, "Entering polycParameters\n"));
  CheckName("polycParameters");
  PushEndHandler(JustPop);
  SetStartHandler(polycSemi_Major_Axis);
}

SEHdeclare(polycSemi_Major_Axis)
{
  next_parameter_handler = SEHname(polycSemi_Minor_Axis);
  TryHandler(Semi_Major_Axis);
}

SEHdeclare(polycSemi_Minor_Axis)
{
  next_parameter_handler = SEHname(polycLongitude_Central_Meridian);
  TryHandler(Semi_Minor_Axis);
}

SEHdeclare(polycLongitude_Central_Meridian)
{
  next_parameter_handler = SEHname(Longitude_Central_Meridian);
  TryHandler(Longitude_Central_Meridian);
}

SEHdeclare(Longitude_Central_Meridian)
{
  TRACE((stderr, "Entering Longitude_Central_Meridian\n"));
  CheckName("Longitude_Central_Meridian");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Longitude_Central_Meridian");
}

SEHdeclare(polycLatitude_Origin)
{
  next_parameter_handler = SEHname(polycFalse_Easting);
  TryHandler(Latitude_Origin);
}

SEHdeclare(Latitude_Origin)
{
  TRACE((stderr, "Entering Latitude_Origin\n"));
  CheckName("Latitude_Origin");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Latitude_Origin");
}

SEHdeclare(polycFalse_Easting)
{
  next_parameter_handler = SEHname(polycFalse_Northing);
  TryHandler(False_Easting);
}

SEHdeclare(polycFalse_Northing)
{
  next_parameter_handler = SEHname(NoMore);
  TryHandler(False_Northing);
}

SEHdeclare(TM)
{
  TRACE((stderr, "Entering TM\n"));
  if(NameIs("TM")) {
    PushEndHandler(JustPop);
    check_proj_attributes(atts, "TM", "HE5_GCTP_TM", "Transverse Mercator", "9");
    projcode = 9;
    SetStartHandler(tmParameters);
  }
  else TryHandler(LAMAZ);
}

SEHdeclare(tmParameters)
{
  TRACE((stderr, "Entering tmParameters\n"));
  CheckName("tmParameters");
  PushEndHandler(JustPop);
  SetStartHandler(tmSemi_Major_Axis);
}

SEHdeclare(tmSemi_Major_Axis)
{
  next_parameter_handler = SEHname(tmSemi_Minor_Axis);
  TryHandler(Semi_Major_Axis);
}

SEHdeclare(tmSemi_Minor_Axis)
{
  next_parameter_handler = SEHname(tmCentral_Meridian_Scale_Factor);
  TryHandler(Semi_Minor_Axis);
}

SEHdeclare(tmCentral_Meridian_Scale_Factor)
{
  next_parameter_handler = SEHname(tmLongitude_Central_Meridian);
  TryHandler(Semi_Minor_Axis);
}

SEHdeclare(Central_Meridian_Scale_Factor)
{
  TRACE((stderr, "Entering Central_Meridian_Scale_Factor\n"));
  CheckName("Central_Meridian_Scale_Factor");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Central_Meridian_Scale_Factor");
}

SEHdeclare(tmLongitude_Central_Meridian)
{
  next_parameter_handler = SEHname(tmLatitude_Origin);
  TryHandler(Longitude_Central_Meridian);
}

SEHdeclare(tmLatitude_Origin)
{
  next_parameter_handler = SEHname(tmFalse_Easting);
  TryHandler(Latitude_Origin);
}

SEHdeclare(tmFalse_Easting)
{
  next_parameter_handler = SEHname(tmFalse_Northing);
  TryHandler(False_Easting);
}

SEHdeclare(tmFalse_Northing)
{
  next_parameter_handler = SEHname(NoMore);
  TryHandler(False_Northing);
}

SEHdeclare(LAMAZ)
{
  TRACE((stderr, "Entering LAMAZ\n"));
  if(NameIs("LAMAZ")) {
    PushEndHandler(JustPop);
    check_proj_attributes(atts, "LAMAZ", "HE5_GCTP_LAMAZ", "Transverse Mercator", "11");
    projcode = 11;
    SetStartHandler(lamazParameters);
  }
  else TryHandler(LAMAZ);
}

SEHdeclare(lamazParameters)
{
  TRACE((stderr, "Entering lamazParameters\n"));
  CheckName("lamazParameters");
  PushEndHandler(JustPop);
  SetStartHandler(lamazReference_Sphere_Radius);
}

SEHdeclare(lamazReference_Sphere_Radius)
{
  next_parameter_handler = SEHname(Longitude_Center);
  TryHandler(Reference_Sphere_Radius);
}

SEHdeclare(Reference_Sphere_Radius)
{
  TRACE((stderr, "Entering Reference_Sphere_Radius\n"));
  CheckName("Reference_Sphere_Radius");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Reference_Sphere_Radius");
}

SEHdeclare(Longitude_Center)
{
  TRACE((stderr, "Entering Longitude_Center\n"));
  CheckName("Longitude_Center");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Longitude_Center");
  next_parameter_handler = SEHname(Latitude_Center);
}

SEHdeclare(Latitude_Center)
{
  TRACE((stderr, "Entering Latitude_Center\n"));
  CheckName("Latitude_Center");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Latitude_Center");
  next_parameter_handler = SEHname(lamazFalse_Easting);
}

SEHdeclare(lamazFalse_Easting)
{
  next_parameter_handler = SEHname(lamazFalse_Northing);
  TryHandler(False_Easting);
}

SEHdeclare(lamazFalse_Northing)
{
  next_parameter_handler = SEHname(NoMore);
  TryHandler(False_Northing);
}

SEHdeclare(HOM)
{
  TRACE((stderr, "Entering HOM\n"));
  if(NameIs("HOM")) {
    PushEndHandler(JustPop);
    check_proj_attributes(atts, "HOM", "HE5_GCTP_HOM", "Hotine Oblique Mercator", "20");
    projcode = 20;
    SetStartHandler(hom_aParameters);
  }
  else TryHandler(SOM);
}

SEHdeclare(hom_aParameters)
{
  TRACE((stderr, "Entering hom_aParameters\n"));
  if(NameIs("hom_aParameters")) {
    PushEndHandler(JustPop);
    SetStartHandler(hom_aSemi_Major_Axis);
  }
  else TryHandler(hom_bParameters);
}

SEHdeclare(hom_bParameters)
{
  TRACE((stderr, "Entering hom_bParameters\n"));
  if(NameIs("hom_bParameters")) {
    PushEndHandler(JustPop);
    SetStartHandler(hom_bSemi_Major_Axis);
  }
  else {
    fprintf(stderr, "Unrecognized element in HOM: %s\n", name);
    PrintXMLLocation();
    exit(-1);
  }
}

SEHdeclare(hom_aSemi_Major_Axis)
{
  next_parameter_handler = SEHname(hom_aSemi_Minor_Axis);
  TryHandler(Semi_Major_Axis);
}

SEHdeclare(hom_aSemi_Minor_Axis)
{
  next_parameter_handler = SEHname(hom_aCentral_Meridian_Scale_Factor);
  TryHandler(Semi_Minor_Axis);
}

SEHdeclare(hom_aCentral_Meridian_Scale_Factor)
{
  next_parameter_handler = SEHname(hom_aLatitude_Origin);
  TryHandler(Central_Meridian_Scale_Factor);
}

SEHdeclare(hom_aLatitude_Origin)
{
  next_parameter_handler = SEHname(hom_aFalse_Easting);
  TryHandler(Latitude_Origin);
}

SEHdeclare(hom_aFalse_Easting)
{
  next_parameter_handler = SEHname(hom_aFalse_Northing);
  TryHandler(False_Easting);
}

SEHdeclare(hom_aFalse_Northing)
{
  next_parameter_handler = SEHname(Longitude_Center_Line1);
  TryHandler(False_Northing);
}

SEHdeclare(Longitude_Center_Line1)
{
  TRACE((stderr, "Entering Longitude_Center_Line1\n"));
  CheckName("Longitude_Center_Line1");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Longitude_Center_Line1");
  next_parameter_handler = SEHname(Latitude_Center_Line1);
}

SEHdeclare(Latitude_Center_Line1)
{
  TRACE((stderr, "Entering Latitude_Center_Line1\n"));
  CheckName("Latitude_Center_Line1");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Latitude_Center_Line1");
  next_parameter_handler = SEHname(Longitude_Center_Line2);
}

SEHdeclare(Longitude_Center_Line2)
{
  TRACE((stderr, "Entering Longitude_Center_Line2\n"));
  CheckName("Longitude_Center_Line2");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Longitude_Center_Line2");
  next_parameter_handler = SEHname(Latitude_Center_Line2);
}

SEHdeclare(Latitude_Center_Line2)
{
  TRACE((stderr, "Entering Latitude_Center_Line2\n"));
  CheckName("Latitude_Center_Line2");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Latitude_Center_Line2");
  next_parameter_handler = SEHname(hom_aformat_A);
}

SEHdeclare(hom_aformat_A)
{
  next_parameter_handler = SEHname(NoMore);
  TryHandler(format_A);
}

SEHdeclare(format_A)
{
  TRACE((stderr, "Entering format_A\n"));
  CheckName("format_A");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("format_A");
}

SEHdeclare(hom_bSemi_Major_Axis)
{
  next_parameter_handler = SEHname(hom_bSemi_Minor_Axis);
  TryHandler(Semi_Major_Axis);
}

SEHdeclare(hom_bSemi_Minor_Axis)
{
  next_parameter_handler = SEHname(hom_bCentral_Meridian_Scale_Factor);
  TryHandler(Semi_Minor_Axis);
}

SEHdeclare(hom_bCentral_Meridian_Scale_Factor)
{
  next_parameter_handler = SEHname(Azimuth_Angle);
  TryHandler(Central_Meridian_Scale_Factor);
}

SEHdeclare(Azimuth_Angle)
{
  TRACE((stderr, "Entering Azimuth_Angle\n"));
  CheckName("Azimuth_Angle");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Azimuth_Angle");
  next_parameter_handler = SEHname(Azimuth_Longitude);
}

SEHdeclare(Azimuth_Longitude)
{
  TRACE((stderr, "Entering Azimuth_Longitude\n"));
  CheckName("Azimuth_Longitude");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Azimuth_Longitude");
  next_parameter_handler = SEHname(hom_bLatitude_Origin);
}

SEHdeclare(hom_bLatitude_Origin)
{
  next_parameter_handler = SEHname(hom_bFalse_Easting);
  TryHandler(Latitude_Origin);
}

SEHdeclare(hom_bFalse_Easting)
{
  next_parameter_handler = SEHname(hom_bFalse_Northing);
  TryHandler(False_Easting);
}

SEHdeclare(hom_bFalse_Northing)
{
  next_parameter_handler = SEHname(hom_bformat_B);
  TryHandler(False_Northing);
}

SEHdeclare(hom_bformat_B)
{
  next_parameter_handler = SEHname(NoMore);
  TryHandler(format_B);
}

SEHdeclare(format_B)
{
  TRACE((stderr, "Entering format_B\n"));
  CheckName("format_B");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("format_B");
  next_parameter_handler = SEHname(NoMore);
}

SEHdeclare(SOM)
{
  TRACE((stderr, "Entering SOM\n"));
  if(NameIs("SOM")) {
    PushEndHandler(JustPop);
    check_proj_attributes(atts, "SOM", "HE5_GCTP_SOM", "Space Oblique Mercator", "22");
    projcode = 22;
    SetStartHandler(som_aParameters);
  }
  else TryHandler(SOM);
}

SEHdeclare(som_aParameters)
{
  TRACE((stderr, "Entering som_aParameters\n"));
  if(NameIs("som_aParameters")) {
    PushEndHandler(JustPop);
    SetStartHandler(som_aSemi_Major_Axis);
  }
  else TryHandler(som_bParameters);
}

SEHdeclare(som_bParameters)
{
  TRACE((stderr, "Entering som_bParameters\n"));
  if(NameIs("som_bParameters")) {
    PushEndHandler(JustPop);
    SetStartHandler(som_bSemi_Major_Axis);
  }
  else {
    fprintf(stderr, "Unrecognized element in SOM: %s\n", name);
    PrintXMLLocation();
    exit(-1);
  }
}

SEHdeclare(som_aSemi_Major_Axis)
{
  next_parameter_handler = SEHname(som_aSemi_Minor_Axis);
  TryHandler(Semi_Major_Axis);
}

SEHdeclare(som_aSemi_Minor_Axis)
{
  next_parameter_handler = SEHname(Orbit_Inclination);
  TryHandler(Semi_Minor_Axis);
}

SEHdeclare(Orbit_Inclination)
{
  TRACE((stderr, "Entering Orbit_Inclination\n"));
  CheckName("Orbit_Inclination");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Orbit_Inclination");
  next_parameter_handler = SEHname(Longitude_Ascending_Orbit);
}

SEHdeclare(Longitude_Ascending_Orbit)
{
  TRACE((stderr, "Entering Longitude_Ascending_Orbit\n"));
  CheckName("Longitude_Ascending_Orbit");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Longitude_Ascending_Orbit");
  next_parameter_handler = SEHname(som_aFalse_Easting);
}

SEHdeclare(som_aFalse_Easting)
{
  next_parameter_handler = SEHname(som_aFalse_Northing);
  TryHandler(False_Easting);
}

SEHdeclare(som_aFalse_Northing)
{
  next_parameter_handler = SEHname(Period_of_Revolution);
  TryHandler(False_Northing);
}

SEHdeclare(Period_of_Revolution)
{
  TRACE((stderr, "Entering Period_of_Revolution\n"));
  CheckName("Period_of_Revolution");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Period_of_Revolution");
  next_parameter_handler = SEHname(Satellite_Ratio);
}

SEHdeclare(Satellite_Ratio)
{
  TRACE((stderr, "Entering Satellite_Ratio\n"));
  CheckName("Satellite_Ratio");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Satellite_Ratio");
  next_parameter_handler = SEHname(Landsat_End_Of_Path);
}

SEHdeclare(Landsat_End_Of_Path)
{
  TRACE((stderr, "Entering Landsat_End_Of_Path\n"));
  CheckName("Landsat_End_Of_Path");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Landsat_End_Of_Path");
  next_parameter_handler = SEHname(som_aHDF_EOS_Variable);
}

SEHdeclare(som_aHDF_EOS_Variable)
{
  next_parameter_handler = SEHname(som_aformat_A);
  TryHandler(HDF_EOS_Variable);
}

SEHdeclare(HDF_EOS_Variable)
{
  TRACE((stderr, "Entering HDF-EOS_Variable\n"));
  CheckName("HDF-EOS_Variable");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("HDF-EOS_Variable");
}

SEHdeclare(som_aformat_A)
{
  next_parameter_handler = SEHname(NoMore);
  TryHandler(format_A);
}

SEHdeclare(som_bSemi_Major_Axis)
{
  next_parameter_handler = SEHname(som_bSemi_Minor_Axis);
  TryHandler(Semi_Major_Axis);
}

SEHdeclare(som_bSemi_Minor_Axis)
{
  next_parameter_handler = SEHname(Landsat_Satellite_Number);
  TryHandler(Semi_Minor_Axis);
}

SEHdeclare(Landsat_Satellite_Number)
{
  TRACE((stderr, "Entering Landsat_Satellite_Number\n"));
  CheckName("Landsat_Satellite_Number");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Landsat_Satellite_Number");
  next_parameter_handler = SEHname(Landsat_Path_Number);
}

SEHdeclare(Landsat_Path_Number)
{
  TRACE((stderr, "Entering Landsat_Path_Number\n"));
  CheckName("Landsat_Path_Number");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Landsat_Path_Number");
  next_parameter_handler = SEHname(som_bFalse_Easting);
}

SEHdeclare(som_bFalse_Easting)
{
  next_parameter_handler = SEHname(som_bFalse_Northing);
  TryHandler(False_Easting);
}

SEHdeclare(som_bFalse_Northing)
{
  next_parameter_handler = SEHname(som_bHDF_EOS_Variable);
  TryHandler(False_Northing);
}

SEHdeclare(som_bHDF_EOS_Variable)
{
  next_parameter_handler = SEHname(som_bformat_B);
  TryHandler(HDF_EOS_Variable);
}

SEHdeclare(som_bformat_B)
{
  next_parameter_handler = SEHname(NoMore);
  TryHandler(format_B);
}

SEHdeclare(GOOD)
{
  TRACE((stderr, "Entering GOOD\n"));
  if(NameIs("GOOD")) {
    PushEndHandler(JustPop);
    check_proj_attributes(atts, "GOOD", "HE5_GCTP_GOOD", "Interrupted Goode Homolosine", "24");
    projcode = 24;
    SetStartHandler(goodParameters);
  }
  else TryHandler(ISINUS);
}

SEHdeclare(goodParameters)
{
  TRACE((stderr, "Entering goodParameters\n"));
  CheckName("goodParameters");
  PushEndHandler(JustPop);
  SetStartHandler(goodReference_Sphere_Radius);
}

SEHdeclare(goodReference_Sphere_Radius)
{
  next_parameter_handler = SEHname(NoMore);
  TryHandler(Reference_Sphere_Radius);
}

SEHdeclare(ISINUS)
{
  TRACE((stderr, "Entering ISINUS\n"));
  if(NameIs("ISINUS")) {
    PushEndHandler(JustPop);
    check_proj_attributes(atts, "ISINUS", "HE5_GCTP_ISINUS", "Interrupted Sinusoidal Projection", "31");
    projcode = 31;
    SetStartHandler(isinusParameters);
  }
  else {
    fprintf(stderr, "Unknown element inside Projection: %s\n", name);
    PrintXMLLocation();
    exit(-1);
  }
}

SEHdeclare(isinusParameters)
{
  TRACE((stderr, "Entering isinusParameters\n"));
  CheckName("isinusParameters");
  PushEndHandler(JustPop);
  SetStartHandler(isinusReference_Sphere_Radius);
}

SEHdeclare(isinusReference_Sphere_Radius)
{
  next_parameter_handler = SEHname(isinusLongitude_Central_Meridian);
  TryHandler(Reference_Sphere_Radius);
}

SEHdeclare(isinusLongitude_Central_Meridian)
{
  next_parameter_handler = SEHname(isinusFalse_Easting);
  TryHandler(Longitude_Central_Meridian);
}

SEHdeclare(isinusFalse_Easting)
{
  next_parameter_handler = SEHname(isinusFalse_Northing);
  TryHandler(False_Easting);
}

SEHdeclare(isinusFalse_Northing)
{
  next_parameter_handler = SEHname(Latitudinal_Zones);
  TryHandler(False_Northing);
}

SEHdeclare(Latitudinal_Zones)
{
  TRACE((stderr, "Entering Latitudinal_Zones\n"));
  CheckName("Latitudinal_Zones");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Latitudinal_Zones");
  next_parameter_handler = SEHname(Right_Justified_Columns);
}

SEHdeclare(Right_Justified_Columns)
{
  TRACE((stderr, "Entering Right_Justified_Columns\n"));
  CheckName("Right_Justified_Columns");
  PushEndHandler(StuffParameter);
  StartCollectingCDATA();
  GrabParamIndex("Right_Justified_Columns");
  next_parameter_handler = SEHname(NoMore);
}

EEHdeclare(Projection)
{
  PopHandler();
  SetStartHandler(SphereCode);
}

SEHdeclare(SphereCode)
{
  TRACE((stderr, "Entering SphereCode\n"));
  CheckName("SphereCode");
  PushEndHandler(SphereCode);
  StartCollectingCDATA();
}

typedef struct codemap {
  char const *name;
  int code;
} codemap;

codemap sphere_codes[21] = {
  {"Clarke 1866", 0}, {"Clarke 1880", 1}, {"Bessel", 2},
  {"International 1967", 3}, {"International 1909", 4}, {"WGS 72", 5},
  {"Everest", 6}, {"WGS 66", 7}, {"GRS 1980", 8}, {"Airy", 9},
  {"Modified Airy", 10}, {"Modified Everest", 11}, {"WGS 84", 12},
  {"Southeast Asia", 13}, {"Austrailian National", 14}, {"Krassovsky", 15},
  {"Hough", 16}, {"Mercury 1960", 17}, {"Modified Mercury 1968", 18},
  {"Sphere of radius 637099 meters", 19}, {"NONE", 0}};


EEHdeclare(SphereCode)
{
  char const *sstring;
  /* int i; */

  PopHandler();
  StopCollectingCDATA();
  sstring = GlobalCDATA.buffer;
  /* I'm going to leave this here and replace it with a simple
     read-the-number hack */
#if 0
  spherecode = -1;
  for(i=0; i< sizeof(sphere_codes)/sizeof(codemap); i++)
    if(!strcmp(sstring, sphere_codes[i].name)) {
      spherecode = sphere_codes[i].code;
      break;
    }

  if(spherecode == -1) {
    fprintf(stderr, "Unknown Sphere code: %s\n", sstring);
    PrintXMLLocation();
    exit(-1);
  }
#endif
  if(sscanf(sstring, "%d", &spherecode) != 1) {
    fprintf(stderr, "Unknown Sphere code: %s\n", sstring);
    PrintXMLLocation();
    exit(-1);
  }
  SetStartHandler(ZoneCode);
}

SEHdeclare(ZoneCode)
{
  TRACE((stderr, "Entering ZoneCode\n"));
  CheckName("ZoneCode");
  PushEndHandler(ZoneCode);
  StartCollectingCDATA();
}

codemap zone_codes[61] = {
  {"177W", 1},{"171W", 2},{"165W", 3},{"159W", 4},{"153W", 5},{"147W", 6},
  {"141W", 7},{"135W", 8},{"129W", 9},{"123W", 10},{"117W", 11},{"111W", 12},
  {"105W", 13},{"099W", 14},{"093W", 15},{"087W", 16},{"081W", 17},{"075W", 18},
  {"069W", 19},{"063W", 20},{"057W", 21},{"051W", 22},{"045W", 23},{"039W", 24},
  {"033W", 25},{"027W", 26},{"021W", 27},{"015W", 28},{"009W", 29},{"003W", 30},
  {"003E", 31},{"009E", 32},{"015E", 33},{"021E", 34},{"027E", 35},{"033E", 36},
  {"039E", 37},{"045E", 38},{"051E", 39},{"057E", 40},{"063E", 41},{"069E", 42},
  {"075E", 43},{"081E", 44},{"087E", 45},{"093E", 46},{"099E", 47},{"105E", 48},
  {"111E", 49},{"117E", 50},{"123E", 51},{"129E", 52},{"135E", 53},{"141E", 54},
  {"147E", 55},{"153E", 56},{"159E", 57},{"165E", 58},{"171E", 59},{"177E", 60},
  {"NONE", 0}
};


EEHdeclare(ZoneCode)
{
  char const *sstring;
  /* int i; */

  PopHandler();
  StopCollectingCDATA();
  sstring = GlobalCDATA.buffer;

  /* I'm going to leave this here, commented out, and replace it
     with a simple read-the-number hack */
#if 0
  zonecode = -1;
  for(i=0; i< sizeof(zone_codes)/sizeof(codemap); i++)
    if(!strcmp(sstring, zone_codes[i].name)) {
      zonecode = zone_codes[i].code;
      break;
    }

  if(zonecode == -1) {
    fprintf(stderr, "Unknown Zone code: %s\n", sstring);
    PrintXMLLocation();
    exit(-1);
  }
#endif
  if(sscanf(sstring, "%d", &zonecode) != 1) {
    fprintf(stderr, "Unknown Sphere code: %s\n", sstring);
    PrintXMLLocation();
    exit(-1);
  }
  SetStartHandler(PixelRegistration);
}

SEHdeclare(PixelRegistration)
{
  XML_Char const *prstring = FindAttribute("Code");
  TRACE((stderr, "Entering PixelRegistration\n"));
  CheckName("PixelRegistration");
  PushEndHandler(JustPop);
  if(prstring == NULL) {
    fprintf(stderr, "PixelRegistration with no Code attribute\n");
    PrintXMLLocation();
    exit(-1);
  }

  if(!strcmp(prstring, "HE5_HDFE_CENTER"))
    prcode = 0;
  else if (!strcmp(prstring, "HE5_HDFE_CORNER"))
    prcode = 1;
  else {
    fprintf(stderr, "Unrecognized PixelRegistration Code attribute: %s\n", prstring);
    PrintXMLLocation();
    exit(-1);
  }
  SetStartHandler(OriginType);
}

SEHdeclare(OriginType)
{
  XML_Char const *otstring = FindAttribute("Code");
  TRACE((stderr, "Entering OriginType\n"));
  CheckName("OriginType");
  PushEndHandler(JustPop);
  if(otstring == NULL) {
    fprintf(stderr, "OriginType with no Code attribute\n");
    PrintXMLLocation();
    exit(-1);
  }

  if(!strcmp(otstring, "HE5_HDFE_GD_UL"))
    origincode = 0;
  else if (!strcmp(otstring, "HE5_HDFE_GD_UR"))
    origincode = 1;
  else if (!strcmp(otstring, "HE5_HDFE_GD_LL"))
    origincode = 2;
  else if (!strcmp(otstring, "HE5_HDFE_GD_LR"))
    origincode = 3;
  else {
    fprintf(stderr, "Unrecognized OriginType Code attribute: %s\n", otstring);
    PrintXMLLocation();
    exit(-1);
  }
  SetStartHandler(NoMore);
}


SEHdeclare(GridDimension)
{
  TRACE((stderr, "Entering GridDimension\n"));
  if(NameIs("Dimension")) {
    PushEndHandler(GridDimension);
    TryHandler(Dimension);
  }
  else
    TryHandler(GridDataField);
}

EEHdeclare(GridDimension)
{
  
  PopHandler();
  TRACE((stderr, "Exiting GridDimension\n"));

  /* Snarf/assert the Dimension here */
  TRACE((stderr, "Creating grid dimension %s of size %ld\n",
	 dimension_name, (long int)dimension_size));
  SS(HE5_GDdefdim(grid_handle, dimension_name, dimension_size));
  free_safely(dimension_name);

  /* Handling the loop */
  SetStartHandler(GridDimension);
}

/* <!ELEMENT GridDataField (Attribute*, DataType, DimList, FieldStorage?, TileDimList?, link_to_data)> */

SEHdeclare(GridDataField)
{
  TRACE((stderr, "Entering GridDataField\n"));
  if(NameIs("GridDataField")) {
    XML_Char const *field_name = FindAttribute("Name");

    if(field_name == NULL) {
      fprintf(stderr, "GridDataField with no Name attribute\n");
      PrintXMLLocation();
      exit(-1);
    }
    grid_field_name = strclone(field_name);

    PushEndHandler(GridDataField);
    SetStartHandler(GridDataAttribute);
    field_attributes = NULL;
  }
  else
    TryHandler(DataField);
}

EEHdeclare(GridDataField)
{
  PopHandler();

  /* close/release grid data field here */
  free_safely(grid_field_name);
}


SEHdeclare(GridDataFieldStorage)
{
  /* This is where we finally have all the data we need to define the field,
     unroll the attributes, etc.  We'll do it now before we look at the
     FieldStorage?, TileDimList?, link_to_data part.
  */
  SS(HE5_GDdeffield(grid_handle, grid_field_name, griddimnames, gridmaxdimnames,
		    datatype_native_type, fieldstorage_mergecode));

  /* According to the HDF-EOS documentation, we must detach/attach to
     the swath after creating a field and before writing to it */
  SS(HE5_GDdetach(grid_handle));
  SS(grid_handle = HE5_GDattach(hdfeos_handle, grid_name));

  /* Unroll the buffered attributes and write them out.*/

  while(field_attributes != NULL) {
    attributestack *as = field_attributes;
    SS(HE5_GDwritelocattr(grid_handle, grid_field_name,
			  as->name, as->type, &(as->count), as->buffer));
    free_safely(as->name);
    free_safely(as->buffer);
    field_attributes = field_attributes->next;
    free_safely(as);
  }

  if(NameIs("FieldStorage")) {
    PushEndHandler(GridDataFieldStorage);
    TryHandler(FieldStorage);
  }
  else
    TryHandler(TileDimList);
}

EEHdeclare(GridDataFieldStorage)
{
  PopHandler();
  /* This is disabled until the API can actually handle it...
  SS(HE5_SWdefcomp(swath_handle, fieldstorage_compcode, fieldstorage_compparms));
  */
  if(fieldstorage_compparms != NULL)
    free_safely(fieldstorage_compparms);
  SetStartHandler(TileDimList);
}

int tilecode;
int tilerank;
hsize_t *tiledims;
int tiledimcount;

SEHdeclare(TileDimList)
{
  if(NameIs("TileDimList")) {
    XML_Char const *tdcode = FindAttribute("Code");
    XML_Char const *tdrank = FindAttribute("Rank");

    TRACE((stderr, "Entering TileDimList\n"));
    if(tdcode == NULL) {
      fprintf(stderr, "TileDimList with no Code attribute\n");
      PrintXMLLocation();
      exit(-1);
    }
    if(tdrank == NULL) {
      fprintf(stderr, "TileDimList with no Rank attribute\n");
      PrintXMLLocation();
      exit(-1);
    }
    /* parse out the code and rank and set the globals */
    if(!strcmp(tdcode, "HE5_HDFE_NOTILE"))
      tilecode = HE5_HDFE_NOTILE;
    else if(!strcmp(tdcode, "HE5_HDFE_TILE"))
      tilecode = HE5_HDFE_TILE;
    else {
      fprintf(stderr, "TileDimList with invalid Code: %s\n", tdcode);
      PrintXMLLocation();
      exit(-1);
    }
    if(sscanf(tdrank, "%d", &tilerank) != 1) {
      fprintf(stderr, "TileDimList with invalid Rank: %s\n", tdrank);
      PrintXMLLocation();
      exit(-1);
    }
    /* set up to gather up the list of tile dimensions */
    PushEndHandler(TileDimList);
    tiledims = malloc(sizeof(hsize_t) * tilerank);
    tiledimcount = 0;
    SetStartHandler(TileDim);
  }
  else
    TryHandler(GridDatalink_to_data);
}

SEHdeclare(TileDim)
{
  TRACE((stderr, "Entering TileDim\n"));
  CheckName("TileDim");
  PushEndHandler(TileDim);
  StartCollectingCDATA();
}

EEHdeclare(TileDim)
{
  hsize_t tparam;

  PopHandler();
  StopCollectingCDATA();
  if(sscanf(GlobalCDATA.buffer, "%ld", &tparam) != 1) {
    fprintf(stderr, "Not a tile dimension number: %s\n", GlobalCDATA.buffer);
    PrintXMLLocation();
    exit(-1);
  }

  if(tiledimcount >= tilerank) {
    fprintf(stderr, "Too many tile dimensions\n");
    PrintXMLLocation();
    exit(-1);
  }
  tiledims[tiledimcount++] = tparam;
}


EEHdeclare(TileDimList)
{
  /* grab the gathered tile dimensions, do the GDdeftile here */
  if(tilecode != HE5_HDFE_NOTILE)
    SS(HE5_GDdeftile(grid_handle, tilecode, tilerank, tiledims));
  free_safely(tiledims);

  SetStartHandler(GridDatalink_to_data);
}

SEHdeclare(GridDatalink_to_data)
{
  TRACE((stderr, "Entering GridDatalink_to_data\n"));
  if(NameIs("link_to_data")) {
    PushEndHandler(GridDatalink_to_data);
    StartCollectingCDATA();
  }
  else
    TryHandler(GridDatain_line_data);
}

EEHdeclare(GridDatalink_to_data)
{
  void *buffer;
  hssize_t *start_vector = calloc(griddimcount, sizeof(hsize_t));

  PopHandler();
  StopCollectingCDATA();

  /* Dereference the link, hose the data into the field */
  buffer = inhale_file();
  /* We make the following assumption about this data file:
   * - It's the right byte order (it was created on a similar system)
   If this assumption is wrong, the resulting HDF-EOS5 file will be wrong.
  */

  /* Write the *&^%$#@! data already... */
  SS(HE5_GDwritefield(grid_handle, grid_field_name, start_vector, NULL, griddimsizes, buffer));
  
  /* clean up after ourselves */
  free_safely(buffer);
  free_safely(start_vector);

}

void *grid_field_buffer;

SEHdeclare(GridDatain_line_data)
{
  TRACE((stderr, "Entering GridDatain_line_data\n"));
  if(NameIs("in_line_data")) {
    PushEndHandler(GridDatain_line_data);
    if((grid_field_buffer = calloc(datatype_count, datatype_size)) == NULL) {
      fprintf(stderr, "Not enough memory to allocate in_line_data buffer\n");
      PrintXMLLocation();
      exit(-1);
    }
    StartCollectingInLineData(grid_field_buffer, datatype_count, datatype_native_type);
  }
  else {
    fprintf(stderr, "Expected link_to_data or in_line_data, saw %s\n", name);
    PrintXMLLocation();
    exit(-1);
  }
}

EEHdeclare(GridDatain_line_data)
{
  hssize_t *start_vector = calloc(dimnamecount, sizeof(hsize_t));

  TRACE((stderr, "Exiting GridDatain_line_data\n"));
  PopHandler();
  StopCollectingInLineData();

  /* Write the *&^%$#@! data already... */
  SS(HE5_GDwritefield(grid_handle, grid_field_name, start_vector, NULL, griddimsizes, grid_field_buffer));
  
  /* clean up after ourselves */
  free_safely(grid_field_buffer);
  free_safely(start_vector);
}
  
