/* Handles the callbacks for <Swath> and its children */

#include "xml2he5.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include "globals.h"


/* <!ELEMENT Swath (Attribute*, Dimension*, GeoField*, DataField*, DimensionMapping*)> */
SEHdeclare(Swath)
{

  XML_Char const *xc_swath_name = FindAttribute("Name");

  TRACE((stderr, "Entering Swath\n"));
  CheckName("Swath");
  PushEndHandler(Swath);

  if(xc_swath_name == NULL) {
    fprintf(stderr, "Swath with no Name attribute\n");
    PrintXMLLocation();
    exit(-1);
  }
  swath_name = strclone(xc_swath_name);

  /* Create the Swath */
  SS(swath_handle = HE5_SWcreate(hdfeos_handle, swath_name));

  /* Set up to remember any dimensions that get defined */
  recycle_dm(&dimension_map);

  /* Set up to eat any Attributes */
  SetStartHandler(SwathAttribute);
}

EEHdeclare(Swath)
{
  PopHandler();
  TRACE((stderr, "Exiting Swath\n"));

  /* Close the Swath */
  SS(HE5_SWdetach(swath_handle));
  /* clean up after ourselves */
  free_safely(swath_name);
  recycle_dm(&dimension_map);

  /* Handle the next Swath, if any */
  SetStartHandler(Swath);
}

SEHdeclare(SwathDimension)
{
  TRACE((stderr, "Entering SwathDimension\n"));
  if(NameIs("Dimension")) {
    PushEndHandler(SwathDimension);
    TryHandler(Dimension);
  }
  else
    TryHandler(GeoField);
}

EEHdeclare(SwathDimension)
{
  
  PopHandler();
  TRACE((stderr, "Exiting SwathDimension\n"));

  /* Snarf/assert the Dimension here */
  TRACE((stderr, "Creating swath dimension %s of size %ld\n",
	 dimension_name, (long int)dimension_size));
  SS(HE5_SWdefdim(swath_handle, dimension_name, dimension_size));

  add_dimension(&dimension_map, dimension_name, dimension_size);

  free_safely(dimension_name);

  /* Handling the loop */
  SetStartHandler(SwathDimension);
}

/* <!ELEMENT GeoField (Attribute*, DataType, DimList, FieldStorage?, link_to_data)> */
SEHdeclare(GeoField)
{
  TRACE((stderr, "Entering GeoField\n"));
  if(NameIs("GeoField")) {
    XML_Char const *field_name = FindAttribute("Name");

    if(field_name == NULL) {
      fprintf(stderr, "GeoField with no Name attribute\n");
      PrintXMLLocation();
      exit(-1);
    }
    swath_field_name = strclone(field_name);

    PushEndHandler(GeoField);
    SetStartHandler(GeoFieldAttribute);
    field_attributes = NULL;
  }
  else
    TryHandler(DataField);
}

/* locals to hold accumulated dimension names, sizes, and count */
char *gfdimnames = NULL;
hsize_t *gfdimsizes = NULL;
int gfdimcount;
char *gfmaxdimnames = NULL;
hsize_t *gfmaxdimsizes = NULL;
int gfmaxdimcount;

EEHdeclare(GeoField)
{
  PopHandler();
  free_safely(swath_field_name);
  SetStartHandler(GeoField);
}

SEHdeclare(GeoFieldDimList)
{
   TRACE((stderr, "Entering GeoFieldName\n"));
  CheckName("DimList");
  PushEndHandler(GeoFieldDimList);
  dimnamebuffer = NULL;
  dimsizebuffer = NULL;
  dimnamecount = 0;
  SetStartHandler(DimName);
}

EEHdeclare(GeoFieldDimList)
{
  PopHandler();
  gfdimnames = dimnamebuffer;
  gfdimsizes = dimsizebuffer;
  gfdimcount = dimnamecount;

  SetStartHandler(GeoFieldMaxDimList);
}

SEHdeclare(GeoFieldMaxDimList)
{
  TRACE((stderr, "Entering GeoFieldMaxDimList\n"));
  if(NameIs("MaxDimList")) {
    PushEndHandler(GeoFieldMaxDimList);
    dimnamebuffer = NULL;
    dimsizebuffer = NULL;
    dimnamecount = 0;
    SetStartHandler(DimName);
  }
  else
    TryHandler(GeoFieldFieldStorage);
}

EEHdeclare(GeoFieldMaxDimList)
{
  PopHandler();
  gfmaxdimnames = dimnamebuffer;
  gfmaxdimsizes = dimsizebuffer;
  gfmaxdimcount = dimnamecount;

  SetStartHandler(GeoFieldFieldStorage);
}

/* locals to hold accumulated dimension names, sizes, and count */
char *dfdimnames = NULL;
hsize_t *dfdimsizes = NULL;
int dfdimcount;
char *dfmaxdimnames = NULL;
hsize_t *dfmaxdimsizes = NULL;
int dfmaxdimcount;

SEHdeclare(DataFieldDimList)
{
   TRACE((stderr, "Entering DataFieldName\n"));
  CheckName("DimList");
  PushEndHandler(DataFieldDimList);
  dimnamebuffer = NULL;
  dimsizebuffer = NULL;
  dimnamecount = 0;
  SetStartHandler(DimName);
}

EEHdeclare(DataFieldDimList)
{
  PopHandler();
  dfdimnames = dimnamebuffer;
  dfdimsizes = dimsizebuffer;
  dfdimcount = dimnamecount;

  SetStartHandler(DataFieldMaxDimList);
}

SEHdeclare(DataFieldMaxDimList)
{
  TRACE((stderr, "Entering DataFieldMaxDimList\n"));
  if(NameIs("MaxDimList")) {
    PushEndHandler(DataFieldMaxDimList);
    dimnamebuffer = NULL;
    dimsizebuffer = NULL;
    dimnamecount = 0;
    SetStartHandler(DimName);
  }
  else
    TryHandler(DataFieldFieldStorage);
}

EEHdeclare(DataFieldMaxDimList)
{
  PopHandler();
  dfmaxdimnames = dimnamebuffer;
  dfmaxdimsizes = dimsizebuffer;
  dfmaxdimcount = dimnamecount;

  SetStartHandler(DataFieldFieldStorage);
}


SEHdeclare(GridDataDimList)
{
   TRACE((stderr, "Entering GridDataName\n"));
  CheckName("DimList");
  PushEndHandler(GridDataDimList);
  dimnamebuffer = NULL;
  dimsizebuffer = NULL;
  dimnamecount = 0;
  SetStartHandler(DimName);
}

EEHdeclare(GridDataDimList)
{
  PopHandler();
  griddimnames = dimnamebuffer;
  griddimsizes = dimsizebuffer;
  griddimcount = dimnamecount;

  SetStartHandler(GridDataMaxDimList);
}

SEHdeclare(GridDataMaxDimList)
{
  TRACE((stderr, "Entering GridDataMaxDimList\n"));
  if(NameIs("MaxDimList")) {
    PushEndHandler(GridDataMaxDimList);
    dimnamebuffer = NULL;
    dimsizebuffer = NULL;
    dimnamecount = 0;
    SetStartHandler(DimName);
  }
  else
    TryHandler(GridDataFieldStorage);
}

EEHdeclare(GridDataMaxDimList)
{
  PopHandler();
  gridmaxdimnames = dimnamebuffer;
  gridmaxdimsizes = dimsizebuffer;
  gridmaxdimcount = dimnamecount;

  SetStartHandler(GridDataFieldStorage);
}


hsize_t local_dimsize;

SEHdeclare(DimName)
{

  XML_Char const *local_size_string = FindAttribute("Size");

  TRACE((stderr, "Entering DimName\n"));
  CheckName("DimName");
  PushEndHandler(DimName);
  StartCollectingCDATA();
  if(local_size_string != NULL) {
    if(sscanf(local_size_string, "%ld", &local_dimsize) != 1) {
      fprintf(stderr, "DimName with bogus Size attribute: %s\n",
	      local_size_string);
      PrintXMLLocation();
      exit(-1);
    }
  }
  else
    local_dimsize = 0;    
}

EEHdeclare(DimName)
{

  hsize_t found_size;
  char *dimname;

  PopHandler();
  StopCollectingCDATA();
  dimname = GlobalCDATA.buffer;

  /* Look up the name in the dimension map, and spit if it isn't there */
  if((found_size = find_dimension(dimension_map, dimname)) == 0) {
    fprintf(stderr, "DimName not previously defined: %s\n",
	    dimname);
    PrintXMLLocation();
    exit(-1);
  }
  /* if it is there, and the declared size was unlimited, there better
     have been a Size attribute previously
  */
  if(found_size == H5S_UNLIMITED) {
    if(local_dimsize == 0) {
      fprintf(stderr, "DimName %s was declared UNLIMITED with no local Size attribute\n",
	      dimname);
      PrintXMLLocation();
      exit(-1);
    }
    else
      found_size = local_dimsize;
  }

  if(dimnamebuffer == NULL) {
    dimnamebuffer = strclone(dimname);
    dimsizebuffer = calloc(1, sizeof(hsize_t));
    *dimsizebuffer = found_size;
    dimnamecount = 1;
  }
  else {
    /* note new size = size(current-buffer) + size(newname) + comma + null */
    char *newbuffer = malloc(strlen(dimnamebuffer) + strlen(dimname) + 2);
    
    strcpy(newbuffer, dimnamebuffer);
    strcat(newbuffer, ",");
    strcat(newbuffer, dimname);
    free_safely(dimnamebuffer);
    dimnamebuffer = newbuffer;

    dimsizebuffer = realloc(dimsizebuffer, (++dimnamecount * sizeof(hsize_t)));    dimsizebuffer[dimnamecount-1] = found_size;
  }
  /* and set up the loop */
  SetStartHandler(DimName);
}

SEHdeclare(GeoFieldFieldStorage)
{
  /* int fakeparms = 0; */
  /* Smack the fieldstorage data to its defaults, first */
  fieldstorage_mergecode = 0;
  fieldstorage_compcode = 0;
  /* This is disabled until the API can actually handle it...
  SS(HE5_SWdefcomp(swath_handle, HE5_HDFE_COMP_NONE, &fakeparms));
  */

  if(NameIs("FieldStorage")) {
    PushEndHandler(GeoFieldFieldStorage);
    TryHandler(FieldStorage);
  }
  else
    TryHandler(GeoFieldlink_to_data);
}

EEHdeclare(GeoFieldFieldStorage)
{
  PopHandler();
  /* This is disabled until the API can actually handle it...
  SS(HE5_SWdefcomp(swath_handle, fieldstorage_compcode, fieldstorage_compparms));
  */
  if(fieldstorage_compparms != NULL)
    free_safely(fieldstorage_compparms);
  SetStartHandler(GeoFieldlink_to_data);
}

SEHdeclare(DataFieldFieldStorage)
{
  if(NameIs("FieldStorage")) {
    PushEndHandler(DataFieldFieldStorage);
    TryHandler(FieldStorage);
  }
  else
    TryHandler(DataFieldlink_to_data);
}

EEHdeclare(DataFieldFieldStorage)
{
  PopHandler();
  /* This is disabled until the API can actually handle it...
  SS(HE5_SWdefcomp(swath_handle, fieldstorage_compcode, fieldstorage_compparms));
  */
  if(fieldstorage_compparms != NULL)
    free_safely(fieldstorage_compparms);
  SetStartHandler(DataFieldlink_to_data);
}

SEHdeclare(GeoFieldlink_to_data)
{
  TRACE((stderr, "Entering GeoFieldlink_to_data\n"));

  /* here, finally, is where we have enough data to define the field */
  SS(HE5_SWdefgeofield(swath_handle, swath_field_name, gfdimnames,
		       gfmaxdimnames, datatype_native_type, fieldstorage_mergecode));

  /* According to the HDF-EOS documentation, we must detach/attach to
     the swath after creating a field and before writing to it */
  SS(HE5_SWdetach(swath_handle));
  SS(swath_handle = HE5_SWattach(hdfeos_handle, swath_name));

  /* I think this is where we should unroll the buffered attributes
     and write them out.  The stack nature of the buffer means the
     attributes get written out "backwards", but what the heck...
  */
  while(field_attributes != NULL) {
    attributestack *as = field_attributes;
    SS(HE5_SWwritelocattr(swath_handle, swath_field_name,
			  as->name, as->type, &(as->count), as->buffer));
    free_safely(as->name);
    free_safely(as->buffer);
    field_attributes = field_attributes->next;
    free_safely(as);
  }

  if(NameIs("link_to_data")) {
    PushEndHandler(GeoFieldlink_to_data);
    StartCollectingCDATA();
  }
  else
    TryHandler(GeoFieldin_line_data);
}

void *swath_field_buffer;

SEHdeclare(GeoFieldin_line_data)
{

  TRACE((stderr, "Entering GeoFieldin_line_data\n"));
  if(NameIs("in_line_data")) {
    int i;
    hsize_t total_count = 1;
    /* compute the total elements in the field */
    for(i=0; i<dimnamecount; i++)
      total_count *= dimsizebuffer[i];
    PushEndHandler(GeoFieldin_line_data);
    if((swath_field_buffer = calloc(total_count, datatype_size)) == NULL) {
      fprintf(stderr, "Not enough memory to allocate in_line_data buffer\n");
      PrintXMLLocation();
      exit(-1);
    }
    StartCollectingInLineData(swath_field_buffer, total_count, datatype_native_type);
  }
  else {
    fprintf(stderr, "Expected link_to_data or in_line_data, saw %s\n", name);
    PrintXMLLocation();
    exit(-1);
  }
}

EEHdeclare(GeoFieldin_line_data)
{
  hssize_t *start_vector = calloc(dimnamecount, sizeof(hsize_t));

  TRACE((stderr, "Exiting GeoFieldin_line_data\n"));
  PopHandler();
  StopCollectingInLineData();

  /* Write the *&^%$#@! data already... */
  SS(HE5_SWwritefield(swath_handle, swath_field_name, start_vector, NULL, gfdimsizes, swath_field_buffer));
  
  /* clean up after ourselves */
  free_safely(swath_field_buffer);
  free_safely(start_vector);
}
  
/* opens the file named by the URL in GlobalCDATA.buffer,
   allocates enough memory to hold its entire contents,
   and reads it in.  It might be more elegant to mmap() the file,
   but it would be less portable.
*/
/* What we go through to shut the compiler up ... */
void *inhale_file(void);

void *inhale_file(void)
{
  struct stat filedata;
  int result;
  void *buffer;
  int datahandle;
  char const *filepath = GlobalCDATA.buffer;

  /* GlobalCDATA.buffer must be a file: URL.  The spec for file: URLs
     reads like this:

     fileurl        = "file://" [ host | "localhost" ] "/" fpath

     fpath          = fsegment *[ "/" fsegment ]
     fsegment       = *[ uchar | "?" | ":" | "@" | "&" | "=" ]

     However, damn near everything on the Internet accepts file: URLs
     of the form "file:raw/path/to.doc", so we will as well.

     *sigh*  Sometimes I feel like I'm the only person writing code
     that finds/reads/honors specs any more...
  */
  if(!strncmp(filepath, "file://", 7)) {
    filepath = strchr(filepath + 7, '/');
    if(filepath == NULL) {
      fprintf(stderr, "Bogus file: URL - %s\n", GlobalCDATA.buffer);
      exit(-1);
    }
  }
  else if(!strncmp(filepath, "file:", 5)) {
    filepath += 5;
  }
  else {
    fprintf(stderr, "Bogus file: URL - %s\n", GlobalCDATA.buffer);
    exit(-1);
  }
  /* We make the following assumption about this data file:
   * - It's the right byte order (it was created on a similar system)
   If this assumption is wrong, the resulting HDF-EOS5 file will be wrong.
  */

  if((result = stat(filepath, &filedata)) != 0) {
    perror(filepath);
    exit(-1);
  }
  if((buffer = malloc(filedata.st_size)) == NULL) {
    perror("Error while malloc'ing buffer space");
    exit(-1);
  }
  if((datahandle = open(filepath, O_RDONLY)) < 0) {
    perror(filepath);
    exit(-1);
  }
  if(read(datahandle, buffer, filedata.st_size) != filedata.st_size) {
    perror(filepath);
    exit(-1);
  }
  close(datahandle);
  return(buffer);
}

EEHdeclare(GeoFieldlink_to_data)
{
  void *buffer;
  hssize_t *start_vector = calloc(dimnamecount, sizeof(hsize_t));

  PopHandler();
  StopCollectingCDATA();

  /* Dereference the link, hose the data into the field */
  buffer = inhale_file();
  /* We make the following assumption about this data file:
   * - It's the right byte order (it was created on a similar system)
   If this assumption is wrong, the resulting HDF-EOS5 file will be wrong.
  */

  /* Write the *&^%$#@! data already... */
  SS(HE5_SWwritefield(swath_handle, swath_field_name, start_vector, NULL, gfdimsizes, buffer));
  
  /* clean up after ourselves */
  free_safely(buffer);
  free_safely(start_vector);

}

SEHdeclare(DataFieldlink_to_data)
{
  TRACE((stderr, "Entering DataFieldlink_to_data\n"));

  /* here, finally, is where we have enough data to define the field */
  SS(HE5_SWdefdatafield(swath_handle, swath_field_name, dfdimnames,
			NULL, datatype_native_type, fieldstorage_mergecode));

  /* According to the HDF-EOS documentation, we must detach/attach to
     the swath after creating a field and before writing to it */
  SS(HE5_SWdetach(swath_handle));
  SS(swath_handle = HE5_SWattach(hdfeos_handle, swath_name));

  /* I think this is where we should unroll the buffered attributes
     and write them out.  The stack nature of the buffer means the
     attributes get written out "backwards", but what the heck...
  */
  while(field_attributes != NULL) {
    attributestack *as = field_attributes;
    SS(HE5_SWwritelocattr(swath_handle, swath_field_name,
			  as->name, as->type, &(as->count), as->buffer));
    free_safely(as->name);
    free_safely(as->buffer);
    field_attributes = field_attributes->next;
    free_safely(as);
  }

  if(NameIs("link_to_data")) {
    PushEndHandler(DataFieldlink_to_data);
    StartCollectingCDATA();
  }
  else
    TryHandler(DataFieldin_line_data);
}


SEHdeclare(DataFieldin_line_data)
{
  TRACE((stderr, "Entering DataFieldin_line_data\n"));
  if(NameIs("in_line_data")) {
    int i;
    hsize_t total_count = 1;
    /* compute the total elements in the field */
    for(i=0; i<dimnamecount; i++)
      total_count *= dimsizebuffer[i];
    PushEndHandler(DataFieldin_line_data);
    if((swath_field_buffer = calloc(total_count, datatype_size)) == NULL) {
      fprintf(stderr, "Not enough memory to allocate in_line_data buffer\n");
      PrintXMLLocation();
      exit(-1);
    }
    StartCollectingInLineData(swath_field_buffer, total_count, datatype_native_type);
  }
  else {
    fprintf(stderr, "Expected link_to_data or in_line_data, saw %s\n", name);
    PrintXMLLocation();
    exit(-1);
  }
}

EEHdeclare(DataFieldin_line_data)
{
  hssize_t *start_vector = calloc(dimnamecount, sizeof(hsize_t));

  TRACE((stderr, "Exiting DataFieldin_line_data\n"));
  PopHandler();
  StopCollectingInLineData();

  /* Write the *&^%$#@! data already... */
  SS(HE5_SWwritefield(swath_handle, swath_field_name, start_vector, NULL, dfdimsizes, swath_field_buffer));
  
  /* clean up after ourselves */
  free_safely(swath_field_buffer);
  free_safely(start_vector);
}

EEHdeclare(DataFieldlink_to_data)
{
  void *buffer;
  hssize_t *start_vector = calloc(dimnamecount, sizeof(hsize_t));

  PopHandler();
  StopCollectingCDATA();

  /* Dereference the link, hose the data into the field */
  buffer = inhale_file();
  /* We make the following assumption about this data file:
   * - It's the right byte order (it was created on a similar system)
   If this assumption is wrong, the resulting HDF-EOS5 file will be wrong.
  */

  /* Write the *&^%$#@! data already... */
  SS(HE5_SWwritefield(swath_handle, swath_field_name, start_vector, NULL, dfdimsizes, buffer));
  
  /* clean up after ourselves */
  free_safely(buffer);
  free_safely(start_vector);

}

/* <!ELEMENT DataField (Attribute*, DataFieldName, DataType, DimList, FieldStorage?, link_to_data)> */
SEHdeclare(DataField)
{
  TRACE((stderr, "Entering DataField\n"));
  if(NameIs("DataField")) {
    XML_Char const *field_name = FindAttribute("Name");

    if(field_name == NULL) {
      fprintf(stderr, "DataField with no Name attribute\n");
      PrintXMLLocation();
      exit(-1);
    }
    swath_field_name = strclone(field_name);

    PushEndHandler(DataField);
    SetStartHandler(DataFieldAttribute);
    field_attributes = NULL;
  }
  else
    TryHandler(DimensionMapping);
}

EEHdeclare(DataField)
{
  PopHandler();
  free_safely(swath_field_name);
  SetStartHandler(DataField);
}


/*
<!ELEMENT DimensionMapping (RegularDimensionMapping | IndexDimensionMapping)*>
<!ELEMENT RegularDimensionMapping (GeoDimension, DataDimension, Offset, Increment)>
<!ELEMENT IndexDimensionMapping (GeoDimension, DataDimension, IndexArray)>
*/

SEHdeclare(DimensionMapping)
{
  TRACE((stderr, "Entering DimensionMapping\n"));
  if (NameIs("DimensionMapping")) {
    PushEndHandler(DimensionMapping);
    SetStartHandler(RegularDimensionMapping);
  }
  else {
    fprintf(stderr, "Unexpected element %s in Swath\n", name);
    PrintXMLLocation();
    exit(-1);
  }
}

EEHdeclare(DimensionMapping)
{
  PopHandler();
  SetStartHandler(DimensionMapping);
}

SEHdeclare(RegularDimensionMapping)
{
  TRACE((stderr, "Entering RegularDimensionMapping\n"));
  if(NameIs("RegularDimensionMapping")) {
    PushEndHandler(RegularDimensionMapping);
    SetStartHandler(RDGeoDimension);
  }
  else
    TryHandler(IndexDimensionMapping);
}

EEHdeclare(RegularDimensionMapping)
{
  PopHandler();

  /* Haul stuff out of the globals, define the dmap */
  SS(HE5_SWdefdimmap(swath_handle, geodimension, datadimension,
		     dmoffset, dmincrement));
  /* clean up after ourselves */
  free_safely(geodimension);
  free_safely(datadimension);

  /* And set up the loop */
  SetStartHandler(RegularDimensionMapping);
}

long *indexmapbuffer;

SEHdeclare(IndexDimensionMapping)
{
  TRACE((stderr, "Entering IndexDimensionMapping\n"));
  if(NameIs("IndexDimensionMapping")) {
       PushEndHandler(IndexDimensionMapping);
       SetStartHandler(IDGeoDimension);
  }
  else {
    fprintf(stderr, "Not legal inside DimensionMapping: %s\n", name);
    PrintXMLLocation();
    exit(-1);
  }
}

EEHdeclare(IndexDimensionMapping)
{
  PopHandler();

  /* Haul stuff out of the globals, define the dmap */
  SS(HE5_SWdefidxmap(swath_handle, geodimension, datadimension, indexmapbuffer));
  /* Clean up after ourselves */
  free_safely(geodimension);
  free_safely(datadimension);
  free_safely(indexmapbuffer);

  /* And set up the loop */
  SetStartHandler(RegularDimensionMapping);
}

SEHdeclare(RDGeoDimension)
{
  TRACE((stderr, "Entering RDGeoDimension\n"));
  CheckName("GeoDimension");
  PushEndHandler(RDGeoDimension);
  StartCollectingCDATA();
}

EEHdeclare(RDGeoDimension)
{
  PopHandler();
  StopCollectingCDATA();
  geodimension = strclone(GlobalCDATA.buffer);
  SetStartHandler(RDDataDimension);
}

SEHdeclare(RDDataDimension)
{TRACE((stderr, "Entering RDDataDimension\n"));
  CheckName("DataDimension");
  PushEndHandler(RDDataDimension);
  StartCollectingCDATA();
}

EEHdeclare(RDDataDimension)
{
  PopHandler();
  StopCollectingCDATA();
  datadimension = strclone(GlobalCDATA.buffer);
  SetStartHandler(Offset);
}

SEHdeclare(Offset)
{
  TRACE((stderr, "Entering Offset\n"));
  CheckName("Offset");
  PushEndHandler(Offset);
  StartCollectingCDATA();
}

EEHdeclare(Offset)
{
  PopHandler();
  StopCollectingCDATA();
  if(sscanf(GlobalCDATA.buffer, "%ld", &dmoffset) != 1) {
    fprintf(stderr, "Not a number: %s\n", GlobalCDATA.buffer);
    PrintXMLLocation();
    exit(-1);
  }
  SetStartHandler(Increment);
}

SEHdeclare(Increment)
{
  TRACE((stderr, "Entering Increment\n"));
  CheckName("Increment");
  PushEndHandler(Increment);
  StartCollectingCDATA();
}

EEHdeclare(Increment)
{
  PopHandler();
  StopCollectingCDATA();
  if(sscanf(GlobalCDATA.buffer, "%ld", &dmincrement) != 1) {
    fprintf(stderr, "Not a number: %s\n", GlobalCDATA.buffer);
    PrintXMLLocation();
    exit(-1);
  }
  SetStartHandler(NoMore);
}



SEHdeclare(IDGeoDimension)
{
  TRACE((stderr, "Entering IDGeoDimension\n"));
  CheckName("GeoDimension");
  PushEndHandler(IDGeoDimension);
  StartCollectingCDATA();
}

EEHdeclare(IDGeoDimension)
{
  PopHandler();
  StopCollectingCDATA();
  geodimension = strclone(GlobalCDATA.buffer);
  SetStartHandler(IDDataDimension);
}

SEHdeclare(IDDataDimension)
{
  TRACE((stderr, "Entering IDDataDimension\n"));
  CheckName("DataDimension");
  PushEndHandler(IDDataDimension);
  StartCollectingCDATA();
}

EEHdeclare(IDDataDimension)
{
  PopHandler();
  StopCollectingCDATA();
  datadimension = strclone(GlobalCDATA.buffer);
  SetStartHandler(IndexArray);
}

SEHdeclare(IndexArray)
{
  long gdsize;

  TRACE((stderr, "Entering IndexArray\n"));
  CheckName("IndexArray");
  /* Hmmm... Collect the whole IndexArray how? */
  /* This is going to look a bit like in_line_data... */
  PushEndHandler(IndexArray);
  if((gdsize = find_dimension(dimension_map, geodimension)) == 0) {
    fprintf(stderr, "Undefined dimension: %s\n", geodimension);
    PrintXMLLocation();
    exit(-1);
  }
  indexmapbuffer = calloc(gdsize, sizeof(long));
  StartCollectingInLineData(indexmapbuffer, gdsize, H5T_NATIVE_INT);
}

EEHdeclare(IndexArray)
{
  PopHandler();
  StopCollectingInLineData();
}
