/* The main routine for the HDF-EOS 5 XML reader/file creator */

#include "xml2he5.h"
#include "globals.h"
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#define EXPAT_BUF_SIZE 4096

char *hdfeos_file_name = NULL;

/* The top-level just sets up the inputs and outputs, boots up expat,
   and grinds on the main parse loop until everything is consumed.
*/
/* Command line:
   xml2he5 HDF-EOS-File-Name < XML-FILE
*/
int main(int argc, char **argv)
{

  char expat_buf[EXPAT_BUF_SIZE];
  int docfd = 0; /* reads from stdin */

  /* Random required inits */
  GlobalCDATA.buffer = calloc(1024, 1);

  TheParser = XML_ParserCreate(NULL);
  if(argc != 2) {
    fprintf(stderr, "COMMAND LINE: xml2he5 HDF-EOS-FILE-NAME < XML-FILE-NAME\n");
    exit(-1);
  }
  hdfeos_file_name = argv[1];

  SetStartHandler(HDFEOS_File);

  /* This loop is a slight variation on the code in the expat documentation.
     Unlike that one, mine actually works.
  */
  for(;;) {
    int bytes_read;

    bytes_read = read(docfd, expat_buf, EXPAT_BUF_SIZE);

    if (bytes_read < 0) {
      if(errno != EINTR) {
	perror("xml2he5 read");
	exit(-1);
      }
    }


    if (bytes_read >= 0 &&
	! XML_Parse(TheParser, expat_buf, bytes_read, bytes_read == 0)) {
      PrintXMLLocation();
      fprintf(stderr, "\t%s\n", XML_ErrorString(XML_GetErrorCode(TheParser)));
      exit(-1);
    }

    if (bytes_read == 0)
      break;
  }

  /* everything important is done in the various event handlers,
     so we're done */
  exit(0);
}

/* A stub end handler that just pops.  This ought to be popular ... */
EEHdeclare(JustPop)
{
  PopHandler();
  TRACE((stderr, "Just Popped, stack depth: %d\n", hs_length()));
}

/* A stub start handler that throws a fit.  Can be asserted in places we
   know must not contain any more elements.
*/
SEHdeclare(NoMore)
{
  PrintXMLLocation();
  fprintf(stderr, "Start of element %s found where no elements are allowed\n",
	  name);
  exit(-1);
}

/* <!ELEMENT HDFEOS-File (HDFEOS, HDFEOS_INFORMATION)> */
SEHdeclare(HDFEOS_File)
{    
  TRACE((stderr, "Entering HDFEOS-File\n"));
  CheckName("HDFEOS-File");
  PushEndHandler(HDFEOS_File);
  
  SS(hdfeos_handle = HE5_PTopen(hdfeos_file_name, H5F_ACC_TRUNC));

  SetStartHandler(HDFEOS);
}

EEHdeclare(HDFEOS_File)
{
  TRACE((stderr, "Leaving HDFEOS-File\n"));
  SS(HE5_PTclose(hdfeos_handle));
  PopHandler();
}

/* <!ELEMENT HDFEOS (Attribute*, POINTS?, SWATHS?, GRIDS?)> */
SEHdeclare(HDFEOS)
{
  TRACE((stderr, "Entering HDFEOS\n"));
  CheckName("HDFEOS");
  PushEndHandler(HDFEOS);
  /* Handle name and type attributes here */

  SetStartHandler(FileAttribute);
}

EEHnext(HDFEOS, HDFEOS_INFORMATION)

/* <!ELEMENT HDFEOS_INFORMATION (coremetadata?, archivemetadata?, StructMetadata?, ProductMetadata?)> */
SEHdeclare(HDFEOS_INFORMATION)
{
  TRACE((stderr, "Entering HDFEOS_INFORMATION\n"));
  PushEndHandler(JustPop);
}

/* <!ELEMENT POINTS (Point)+> */
SEHdeclare(POINTS)
{
  if(NameIs("POINTS")) {
    TRACE((stderr, "Entering POINTS\n"));
    PushEndHandler(JustPop);
  }
  else
    TryHandler(SWATHS);
}

/* <!ELEMENT SWATHS (Swath)+> */
SEHdeclare(SWATHS)
{
  if(NameIs("SWATHS")) {
    TRACE((stderr, "Entering SWATHS\n"));
    PushEndHandler(JustPop);

    SetStartHandler(Swath);
  }
  else
    TryHandler(GRIDS);
}

/* <!ELEMENT GRIDS (Grid)+> */
SEHdeclare(GRIDS)
{
  if(NameIs("GRIDS")) {
    TRACE((stderr, "Entering GRIDS\n"));
    PushEndHandler(JustPop);

    SetStartHandler(Grid);
  }
  else {
    fprintf(stderr, "Not legal inside HDFEOS: %s\n", name);
    PrintXMLLocation();
    exit(-1);
  }
}
